/* inaddr-slash-24.c          Version 1.00  2007 February 27
 *
 * Reads a text file with lines of the form:
 *
 * xxx192.42.99.0/24 yyy
 *
 * where xxx can be 0 or more bytes, but if it is not 0 bytes
 * then the character before the first digit of the prefix must
 * be a space.  xxx cannot contain any '/' character.
 *
 * There must be a space after the prefix length.
 *
 * There may be more characters after that.  The address and
 * length numbers may have leading zeroes. Lines should be no
 * longer than 200 bytes.
 *
 * The program does not check for every possible input state and
 * will probably crash if there are not 3 decimal points and the
 * right numbers before the first '/'.
 *
 * For each such line, perform 256 dig requests for an inaddr-arpa
 * lookup of each of the IP addresses in the prefix.  Write the
 * results to a file based on the prefix, with leading zeros:
 *
 *  192.042.099.0-24-inaddr.txt 
 *
 * Compiled with GCC:
 *
 *    gcc inaddr-slash-24.c -o inaddr-slash-24
 *  
 * Copyright public domain - Robin Whittle hhtp://www.firstpr.com.au/ip/
 *
 */



#include <stdio.h>
#include <string.h>

                                    /* Global variables.
                                     */
                                     
long unsigned int loopcount;
                                    /* String to prepare the command line
                                     * for dig in.
                                     */
char digcommandline[80];

                                    /* Input and output file buffers.
                                     */ 
FILE *infile;                                                        
FILE *outfile;                                                       

                                    /* Size of array we load a line of the input
                                     * file into for analysis.
                                     */
#define ARRSIZ 200                  

                                    /* Small array to hold bytes from input file.
                                     */
char arr[ARRSIZ + 1];       
    
                                    /* Byte count of how far we are into the input
                                     * file.
                                     */
int  count;         
                                    /* Pointer to arr[]
                                     */
int  p;             
                                    /* Flag to say we have reached the end of the 
                                     * file, by reading an EOF byte == 0x0A.
                                     */

int eofreached = 0;                                  

                                    /* Set to ignore the current input line.
                                     */
int ignoreline = 0;

                                    /* Flag to break out of file reading loop when
                                     * we reach the end. 0 means don't break.
                                     */
int infiledone = 0;

                                    /* Counter for input lines, starting with 1.
                                     */
int linecount = 0;

                                    /* Loop counter when writing output file.
                                     */
int outloop;



/* Function Prototypes
 *-------------------------------------------------------------------------
 */

void process(void);


main (int argc, char *argv[])
{
                                    /* Check input args and complain if not
                                     * the right number.
                                     */  


    if (argc != 2) 
    {
        printf("Error:                                                                       \n");   
        printf("There must one argument.                                                     \n");
        printf(" inaddr-slash-24 iiiii                                                       \n");
        printf("                                                                             \n");
        printf("   iiiii is the name of the input file.                                      \n");
        
        exit(1);
    }

                                    /* Process the command line arguments.
                                     *
                                     * "r" means open for reading text file.
                                     */
                                     
    if ( ( infile = fopen(argv[1],"r") ) == NULL)
    {
        printf("Can't open %s input file.\n",argv[1]);
        exit(1);
    }                                    
                                 
                    
    while(!feof(infile))
    {
                                        /* Read a line at a time of the input file, 
                                         * into arr[], which we soon feed to the
                                         * process() function to parse.
                                         */
        if(fgets(arr, ARRSIZ, infile))
                                        /* Process the input file and
                                         * write the output file.
                                         */
        {
            process();  
        }
    }
    
    return 0;
}





/*-------------------------------------------------------------------------
 *
 * Processing code - this is what does the interesting work.
 *
 * For this program, we are looking for line containing "/24".  
 *
 * We find it by looking for the first '/' on the line.
 *
 * When we find it we work backwards to find the address and then 
 * announce we found it. Then we perform the DNS lookup, writing 
 * the output to a file. 
 *
 * We keep doing this for all input lines.
 */

void process(void)
{
                                        /* Temp string.
                                         */
    char temp[ARRSIZ + 1];
    
    
                                        /* Pointer to within arr[].
                                         */
    char *arrp;
    
    
                                        /* Pointers to within temp[] where we
                                         * find the start of the four parts
                                         * of the prefix, and the prefix length.
                                         *
                                         * The last one, tempspace, is where we
                                         * find the space after the prefix length.
                                         */
                                         
    char *tempddd, *tempccc, *tempbbb, *tempaaa, *tempnn, *tempspace ;

                                        /* Integers for lengths of the address fields.
                                         */
    int lddd, lccc, lbbb, laaa;

                                        /* Integers for the four bytes of 
                                         * the prefix address - though we 
                                         * don't use pfaaa.
                                         */
    int pfddd, pfccc, pfbbb, pfaaa;                                      

                                        /* The address of the prefix.
                                         */
    int pfaddress;
    
                                        /* How many lines we found which match
                                         * with a valid address and "/24".
                                         */
    int matchinglines = 0;
                                        /* The status which dig returns.
                                         */
    int digexitstatus = 0;
    
    
                                        /* Look for a newline within 199
                                         * characters of start of line.
                                         */
    arrp = strchr(arr, '\n');                                           

                                        /* Set p ready for refill, and exit
                                         * if the line was too long.
                                         * 
                                         * Actually, p was for an old system
                                         * I am not using here . . .
                                         */
                                        
    p = arrp - arr + 1  ;   
    if (p > 199)
    {
        printf("Line length limit of 199 exceeded. Exiting.\n");    
        exit(1);
    }
    
    if (p == 1)
    {
        printf("Last line reached.  %d prefix lines found \n", linecount);
    }
    else
    {   
                                        /* Copy line into temp[] where we can 
                                         * work on it without upsetting arr[].
                                         *
                                         * strncpy() does not put a 0 at the end.
                                         */
        strncpy(temp, arr, p);
        temp[p] =0;



                                        /* Debugging lines
                                         */ 
/*                                                      
 *      printf("Line length = %d\n", p);        
 *      printf("Line is: >>%s<<\n", temp);      
 */ 
                                        /* -----------------------------------
                                         * 
                                         * Look for '/24'.
                                         *
                                         * Make tempnn point to where we expect
                                         * to find the '2'.
                                         */

        tempnn = strchr(temp, '/') + 1;
        if (   (tempnn[0]   != '2')
             ||(tempnn[1]   != '4')
           )
        {
            printf("Expecting \"24\" after first first \'/\'. Exiting. p = %d string = %s\n", p, arr);  
            exit(1);
        }

                                        /* Now look backwards for the dot between bbb and aaa.
                                         * The possible arrangements are:
                                         *
                                         *         tempnn
                                         *              |
                                         *        -54321|
                                         *              |   
                                         *           .0/24
                                         *          .00/24
                                         *         .000/24
                                         *
                                         *  
                                         */
                                         
        if  (tempnn[-2] != '0')
        {
            printf("Expecting at least one zero before the slash. Exiting. p = %d string = %s\n", p, arr);  
            exit(1);
        }

                                        /* Look for 1, 2 or 3 zeroes.
                                         * Set tempaaa to point to the first
                                         * of these 1, 2 or 3 zeroes.
                                         */
                                         
        if      (tempnn[-3] == '.')
        {
            tempaaa = tempnn - 2;
        }
        else if (   (tempnn[-3] == '0')
                  &&(tempnn[-4] == '.') 
                )  
        {
            tempaaa = tempnn - 3;
        }
        else if (   (tempnn[-3] == '0')
                  &&(tempnn[-4] == '0')
                  &&(tempnn[-5] == '.') 
                )  
        {
            tempaaa = tempnn - 4;
        }
        else
        {
            printf("Could not find dot with 1, 2 or 3 zeroes before \"24\". Exiting. p = %d string = %s\n", p, arr);    
            exit(1);
        }
        
                                        /* Play the same game, but not being
                                         * fussy about digits, to find
                                         * the start of bbb.
                                         *
                                         *
                                         *        tempaaa
                                         *              |
                                         *        -54321|
                                         *              |   
                                         *           .b.0
                                         *          .bb.0
                                         *         .bbb.0
                                         *
                                         * Look for "." in three positions.
                                         * Set tempbbb to point to the first
                                         * of these 1, 2 or 3 zeroes.
                                         */
                                         
        if      (tempaaa[-3] == '.')
        {
            tempbbb = tempaaa - 2;
        }
        else if (tempaaa[-4] == '.')
        {
            tempbbb = tempaaa - 3;
        }
        else if (tempaaa[-5] == '.')
        {
            tempbbb = tempaaa - 4;
        }
        else
        {
            printf("Could not find second dot with proper format before before \"24\". Exiting. p = %d string = %s\n", p, arr); 
            exit(1);
        }
        

                                        /* Play the same game, but not being
                                         * fussy about digits, to find
                                         * the start of ccc.
                                         *
                                         *
                                         *        tempbbb
                                         *              |
                                         *        -54321|
                                         *              |   
                                         *           .c.0
                                         *          .cc.0
                                         *         .ccc.0
                                         *
                                         * Look for "." in three positions.
                                         * Set tempccc to point to the first
                                         * of these 1, 2 or 3 zeroes.
                                         */
                                         
        if      (tempbbb[-3] == '.')
        {
            tempccc = tempbbb - 2;
        }
        else if (tempbbb[-4] == '.')
        {
            tempccc = tempbbb - 3;
        }
        else if (tempbbb[-5] == '.')
        {
            tempccc = tempbbb - 4;
        }
        else
        {
            printf("Could not find third dot with proper format before \"24\". Exiting. p = %d string = %s\n", p, arr); 
            exit(1);
        }

                                        /* Play the same game, but not being
                                         * fussy about digits, to find
                                         * the start of ccc.
                                         *
                                         *
                                         *        tempccc    Value of (tempccc - temp)
                                         *              |    if at start of line
                                         *        -54321|
                                         *              |   
                                         *           ~d.0    2
                                         *          ~dd.0    3
                                         *         ~ddd.0    4
                                         *
                                         * Look for "~" in three positions,
                                         * where "~" means either a space
                                         * or the beginning of the line.
                                         *
                                         * Set tempddd to point to the first
                                         * of these 1, 2 or 3 zeroes.
                                         */
                                         
        if      (   ((tempccc - temp) == 2) 
                  ||(tempccc[-3] == ' ')
                )  
        {
            tempddd = tempccc - 2;
        }
        else if (   ((tempccc - temp) == 3) 
                  ||(tempccc[-4] == ' ')
                )  
        {
            tempddd = tempccc - 3;
        }
        else if (   ((tempccc - temp) == 4) 
                  ||(tempccc[-5] == ' ')
                )  
        {
            tempddd = tempccc - 4;
        }
        else
        {
            printf("Could not find space or start of line before third dot with proper format before \"24\". Exiting. p = %d string = %s\n", p, arr);   
            exit(1);
        }
                                        /* Now convert these digit strings into
                                         * integers.  Note this does not check
                                         * for invalid characters like v in:
                                         *
                                         * "12v.3v.v.0/v4 " 
                                         *
                                         * We are not expecting human-generated
                                         * input - just stuff from a BGP routing
                                         * table.
                                         */
        pfddd = atoi(tempddd);
        pfccc = atoi(tempccc);
        pfbbb = atoi(tempbbb);

                                        /* Check the ranges.
                                         */ 

        if (   (pfddd < 0)
             ||(pfddd > 223)
           )
        {
            printf("Expecting ddd in 0 to 223 range. Exiting.\n");  
            exit(1);
        }

        if (   (pfccc < 0)
             ||(pfccc > 255)
           )
        {
            printf("Expecting ccc in 0 to 255 range. Exiting.\n");  
            exit(1);
        }

        if (   (pfbbb < 0)
             ||(pfbbb > 255)
           )
        {
            printf("Expecting bbb in 0 to 255 range. Exiting.\n");  
            exit(1);
        }

        linecount++;
                                        /* Debugging line.
                                         */     
                                                    
        /* printf("Found: ddd=%d  ccc=%d  bbb=%d  linecount=%d p=%d\n", pfddd, pfccc, pfbbb, linecount, p);
         */
                                        /* Display what we are about to do.
                                         */
                                         
        printf("Ready to perform 256 inverse address lookups for %d.%d.%d.0/24  linecount=%d\n", pfddd, pfccc, pfbbb, linecount);
                                         
                                        /* Initially, I fed all the 256 queries to 
                                         * dig in a single file, but there is a problem
                                         * if it can't find a server for this domain.
                                         * It takes maybe 12 seconds to figure this out,
                                         * and the output file was filled with error
                                         * messages.
                                         * 
                                         * So I need to call dig individually, and 
                                         * quit calling it if it returns any message
                                         * such as:
                                         *
                                         *   ;; connection timed out; no servers could be reached   
                                         *
                                         */
        
                                        /* We use dig's "+short" option
                                         * so it only puts out an answer if 
                                         * one is found.  This will be a text name
                                         * which the IP address is reverse mapped to.
                                         *
                                         *  +short -x 12.34.56.1
                                         *
                                         */

        for (loopcount = 0; (loopcount < 256) && (digexitstatus == 0); loopcount++)
        {

/*          printf("loopcount = %d\n", loopcount);  
 */
                                        /* Final output file is of the form:
                                         *
                                         *   192.042.099.0-24-inaddr.txt    
                                         *
                                         * with leading zeros to facilitate
                                         * sorting.
                                         *
                                         *  ddd.ccc.bbb.0-24-inaddr.txt     
                                         *
                                         * The temporary output file is:
                                         * 
                                         *  ddd.ccc.bbb.0-24-inaddr-temp.txt    
                                         *
                                         * This is fed to grep and then deleted.
                                         *
                                    !!   * I had all sorts of trouble with a command line
                                    !!   * which piped the output of dig through grep.
                                    !!   *
                                    !!   * For instance, the loop it was in only ran once!
                                         *
                                         */

                                        /* For the first attempt, we will write to 
                                         * the temp output file, erasing anything which 
                                         * might have been there before.  Subseqent
                                         * runs of dig use >> to append to the file.
                                         *
                                         * This way, we can see it growing, which is
                                         * handy for debugging.
                                         *
                                         * If dig can't find the server, it may take some
                                         * time to return, and the return value will not
                                         * be 0.
                                         *                                       
                                         */


                                         
/*                                         Unfortunately, I find any attempt to run grep
 *                                         with system() causes major malfunctions.
                                         *
                                         * Below is a version without the temp file and
                                         * grep idea.
                                         
                                         
            if (loopcount == 0)
            {
                sprintf(digcommandline, "dig +short -x %d.%d.%d.%d > %03d.%03d.%03d.0-24-inaddr-temp.txt\n",
                                         pfddd, pfccc, pfbbb, loopcount,
                                         pfddd, pfccc, pfbbb);  
                digexitstatus = system(digcommandline);                 
            }
            else


            {
                sprintf(digcommandline, "dig +short -x %d.%d.%d.%d  >> %03d.%03d.%03d.0-24-inaddr-temp.txt\n", 
                                         pfddd, pfccc, pfbbb, loopcount, 
                                         pfddd, pfccc, pfbbb);  
                digexitstatus = system(digcommandline);                 
            }

        }   
                                        /* Now process the temp file through grep to get
                                         * rid of any error messages.
                                         *
                                         * Write the result to the final file.
                                         * Then delete the temp file.
                                         *  /

        sprintf(digcommandline, "grep -v root-servers.net %03d.%03d.%03d.0-24-inaddr-temp.txt > %03d.%03d.%03d.0-24-inaddr.txt\n", 
                                 pfddd, pfccc, pfbbb,  
                                 pfddd, pfccc, pfbbb);  
        system(digcommandline);                 

        sprintf(digcommandline, "rm %03d.%03d.%03d.0-24-inaddr-temp.txt \n", 
                                 pfddd, pfccc, pfbbb);  
        system(digcommandline);                 

*
*/


                                         
                                         
            if (loopcount == 0)
            {
                sprintf(digcommandline, "dig +short -x %d.%d.%d.%d > %03d.%03d.%03d.0-24-inaddr.txt\n",
                                         pfddd, pfccc, pfbbb, loopcount,
                                         pfddd, pfccc, pfbbb);  
                digexitstatus = system(digcommandline);                 
            }
            else

            {
                sprintf(digcommandline, "dig +short -x %d.%d.%d.%d  >> %03d.%03d.%03d.0-24-inaddr.txt\n", 
                                         pfddd, pfccc, pfbbb, loopcount, 
                                         pfddd, pfccc, pfbbb);  
                digexitstatus = system(digcommandline);                 
            }

        }   


    }
}


