/* ping-scanner-mask-multi.c          Version 1.00  2007 February 26
 *
 * (ping-scanner-mask-multi-nt.c is the same, except it does not write total
 *  results to the output file.  See point 8 below.)
 *
 * This program pings random IPv4 addresses according to a mask set by reading
 * a text file of one or more prefixes.  
 *
 * This enables the one program to simultaneously test the ping-responsive-host 
 * density in a number of prefixes, such as 200 chosen from the /24s currently 
 * advertised.
 *
 * The results for each prefix are reported in a line of the output file such as:
 *
 *  13.3333, 00006, 00045, 00005, 69.178.204.0/22 1024 20050615 arin adv
 *
 *  13.3333  % (acknowledgments received / pings sent)
 *    00006  acknowledgments received 
 *    00045  pings sent
 *    00005  line number from input file for this prefix
 *  
 * The rest of the output line is the original prefix line in the input file.
 *      
 * The results file can then be sorted so the low percentage entries are first.
 * Then it is easy to import the text into a spreadsheet and create a bar-graph
 * showing the ping-responsive-host densities as vertical bars, with the lowest to 
 * the highest arranged left to right.  This would be important, for instance, in 
 * understanding the contribution to the low average host density of /24s caused
 * by the large number (35% or so) of such prefixes which have no host responding
 * to pings.
 *
 * ping-scanner-mask-multi.c is based on ping-scanner-mask.c (which is documented
 * below) with these additional functions:
 *
 *  1 - Each line number in the input file, 1 to a maximum of 65536, sets the
 *      corresponding byte or bytes in mask[i] to its line number.
 *
 *  2 - There is an array, with a data structure for each entry, corresponding
 *      to each line number:
 *
 *      a - pingsent - number of pings sent to IP addresses covered by this 
 *          line number's prefix.
 *
 *      b - ackcount - likewise, the number of acknowledgments received.
 *
 *      c - The line number in the input file where this prefix came from.
 *  
 *      d - The text of that line.
 *
 * 3 - Separate counts of pings and acks are kept, and when the program 
 *     finishes, the output file has a line for each input line, containing the 
 *     percentage followed by the ackcount, the pingsent and the text of the 
 *     input line.  
 * 
 * 4 - An optional 3rd parameter tells the program to select not every line
 *     in the input file, but every 2nd, every 3rd etc. - a span of whatever
 *     the parameter is, with values 2 to 64k being meaningful.
 *
 * 5 - An optional 4th parameter sets and offset for the start of reading the
 *     file.  The intention is that these can be used to make the program 
 *     select a sampled subset of prefixes.  Since each instance of the 
 *     program sends about 1 ping a second, or somewhat more if there is an 
 *     acknwoledgement in less than a second, it is desirable to be able
 *     to run multiple instances.  For instance using:
 *
 *          10 0, 10 1, . . . 10 9
 *
 *     as options for ten instances will cause the ten to collectively scan
 *     the prefixes in the whole input file.  Output files can then be
 *     concatenated and the whole task can be accomplished in 1/10 the time
 *     or to a depth of sampling 10 times what would be done by a single
 *     instance.
 *     
 *     Also, it is desirable to run one or more instances with a loosely
 *     sampled subset of prefixes, such as every 40th prefix, evenly 
 *     sampled over the address range, assuming the input file is in
 *     order of address.
 *
 *     On the console, the program displays the last line it reached, but
 *     with span this may not be the last line of the input file.
 *
 *     Also, depending on the offset and the length of the input file, 
 *     perhaps the program will not read any of the lines of the input file.     
 *
 * 6 - When printing the original line in the output file, the program also
 *     prints an indication of whether this prefix spans other smaller
 *     prefixes.  "D00" means it doesn't. "D01" means it spans 2^1 smaller
 *     prefixes.  This only works with /8 to /24 prefix lengths. The program
 *     doesn't work with prefixes longer than /24.
 *
 * 7 - The program only displays the total ping statistics on the console, 
 *     rather than the statistics for every prefix it is scanning.
 *
 * 8 - By commenting out some code, we create a version 
 *     ping-scanner-mask-multi-nt.c which only writes prefix-specific results 
 *     to the output file, with no total results.  Search below for
 *     "ping-scanner-mask-multi-nt.c".  
 *
 * 9 - The output file name is a concatenation of the input file name, plus
 *
 *       -Ssss-Oooo.txt
 *
 *     showing the span and offset values.
 *
 * Example usage
 * -------------
 *
 * Lets say we have just downloaded http://bgp.potaroo.net/ipv4-stats/prefixes.txt
 * which is a list of the whole IPv4 address space, broken into contiguous subnets, 
 * one per line, with notes on whether it is reserved, allocated and not advertised
 * or allocated and advertised.  Suppose we have processed this with grep to get
 * just the /24s which are advertised, into a file prefixes-adv-24.txt.  This has
 * 14779 lines.  We want to run 50 instances of ping-scanner-mask-multi to 
 * send about 1 ping a second each, to a separate sample of these prefixes, where
 * overall we want to test about 500 prefixes.  So we want each instance to test 
 * about 100 prefixes.
 *
 *
 * We could do this by editing the one file into a bunch of 50 line files, but
 * here is a better way:
 *
 *   ping-scanner-mask-multi prefixes-adv-24.txt 86400 147  1 & 
 *   ping-scanner-mask-multi prefixes-adv-24.txt 86400 147  2 & 
 *   ping-scanner-mask-multi prefixes-adv-24.txt 86400 147  3 & 
 *
 *   . . .
 *
 *   ping-scanner-mask-multi prefixes-adv-24.txt 86400 147 50 & 
 *
 * These will run for about a day.  At the end, we concatenate the output 
 * files (and chop out the summary results line, if this is the standard
 * ping-scanner-mask-multi, and not ping-scanner-mask-multi-nt) and
 * sort the lines so they appear in increasing percentage order.
 * Then you have a little file which can be fed to a graphing program
 * to give a distribution curve of host densities.
 *
 * This program can hog memory.  It contains an array of 225 * 256 * 256
 * 16 bit integers = 30 Megabytes.  If you want to run multiple instances
 * in a way which would stretch system memory, you could compile a new
 * version with mask[] made of 8 bit bytes, and with a limit of 254 on 
 * the number of prefixes allowed to be scanned.  That would reduce the
 * RAM usage by 15 Megabytes.  
 *
 * 
 * (The rest of this block of comments is the same as for ping-scanner-mask.c)
 * 
 * Pings will be sent randomly within the allowed time, usually once a second, for 
 * the number of pings specified.  However, since the timing is done by the
 * external ping program, sometimes the pings may be sent faster than this.
 *
 * An example of the format of the input file is:
 *
 * 192.231.221.0/24 256 19921022 arin adv
 * 192.231.233.0/24 256 19921022 arin adv
 * 192.231.253.0/24 256 19921026 arin adv   
 *
 * where all that matters is the lines start with:
 *
 * "ddd.ccc.bbb.0/mm " or "ddd.ccc.bbb.00/mm " or "ddd.ccc.bbb.000/mm " 
 * 
 * where ddd, ccc and aaa are 1 to 3 digits 0 to 255 (0 to 223 for ddd) and where 
 * mm is 1 or 2 digits 1 to 24.  The maximum line length is 200 chars.
 * 
 * There must be a space, and then there can be 0 or more other characters.
 *
 * It is OK for the prefix lengths and addresses not to be on binary boundaries
 * so:
 *
 *   1.2.3.0/24 
 *   1.2.3.1/23
 *
 * would cover:
 *  
 *   1.2.3.0
 *   1.2.3.1
 *   1.2.3.2
 *
 * If one prefix overlaps another, a single error will be reported on the 
 * console for the first occurence, but the program will continue. 
 *
 * In general, with the pseudorandom pinging, there will be repetitive pinging of 
 * those smaller masks where the number of available IP addresses is less than the 
 * total number of pings sent in the available time.
 *
 * The 31 bit PSRNG is seeded on the current time, in seconds, so results will differ 
 * each  time.  Bits 0 to 23 of the PSRNG are used, for these bits of the address.
 * Bits 24 to 31 of the address are done with a counter.
 * 
 * So for a given random number ccc.bbb.aaa, the program will potentially ping
 * all addressses 0.ccc.bbb.aaa to 223.ccc.bbb.aaa, according to the mask.
 *
 * The program counts the number of responses, which arrive in less than a second, 
 * and writes the total to a text file with the same name as the input file, 
 * with "-results.txt" appended to its name.  It also writes the number of pings
 * sent and number of acknowledgements as a  percentage of the number of pings sent.
 *
 * The purpose is to try to estimate the percentage of IP address space which actually has a
 * computer connected to it, for a particular subset of the address space.  For instance
 * by grepping http://bgp.potaroo.net/ipv4-stats/prefixes.txt for all the prefixes of a
 * particular length.
 *
 * Compile with the GCC C compiler:
 * 
 *   gcc ping-scanner-mask-multi.c -o ping-scanner-mask-multi
 *
 * Check the -w or -t option of your ping program - the one which enables it to exit
 * after a certain time.  Some are -w and some are -t.  Below, I assume -w.
 *
 * The ping program may return in less than a second if it gets an ack packet (including
 * an ack from another IP address, such a router en-route which takes it upon itself
 * to acknowlede pings to a now non-existent subnet).  In order to keep the process 
 * going at a reliable rate, such as for a 24 hour survey, it might be possible to 
 * code some timing functions into this program, but I have chosen not to, since
 * I think that would tie up the CPU waiting for a second to elapse, and that this
 * would get in the way of using multiple instances of this program in parallel, due
 * to CPU load.  For instance, we might want to run an instance for every prefix length
 * /8 to /24.
 *
 * Copyright public domain - Robin Whittle hhtp://www.firstpr.com.au/ip/
 *
 * Please use this judiciously.  Folks may be upset about being pinged, 
 * especially if they are the US Department of Defence!
 *
 */


#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include "rand31-park-miller-carta.c"

                                    /* Global variables.
                                     */
                                     
long unsigned int loopcount;
                                    /* Bytes 3, 2, 1 and 0 of address to ping.
                                     */
unsigned int a3, a2, a1, a0;
    
                                    /* Number of pings sent so far.
                                     */
long int pingsent = 0;

                                    /* Number of acknowledgements received so far.
                                     */
long int ackcount = 0;
                                    /* Percentage of pings returned.
                                     */
float percentage = 0;    
                                    /* Raw random number.
                                     */ 
int rawrand;
    
                                    /* Counter for ddd byte cycle.
                                     */
int dddcount =0;                                                                        
                                    
                                    /* String to prepare the command line in.
                                     */
char commandline[80];
    
                                    /* String to prepare the output file
                                     * name in.
                                     */
char outfilename[80];

                                    /* Input and output file buffers.
                                     */ 
FILE *infile;                                                        
FILE *outfile;                                                       


                                    /* Seed for random number generator.
                                     */
long unsigned int randseed = 0;
    
                                    /* Structure for gettimeofday().
                                     */
struct timeval timenow;
                                    /* Absolute time in seconds when the program
                                     * starts, when it should stop and how many 
                                     * seconds it should run for.
                                     */
                                      
int starttime, endtime, runtime; 


                                    /* 225 * 256 * 256 16 bit integers, each of 
                                     * which is set to 0 unless this particular
                                     * /24 is enabled for pinging due to being 
                                     * included in one of more of the prefixes 
                                     * we read in the input file.  Contains line 
                                     * number which contained the prefix which 
                                     * caused this byte to be set to non-zero.
                                     *
                                     * This is set to one /8 longer than the 
                                     * maximum address we allow: 223.255.255.0
                                     * so we don't have to worry about overruns.
                                     *
                                     * 225 * 256 * 256 = 15264000.
                                     */
                                     
#define MASKLEN 15264000
    
short unsigned int mask[MASKLEN];

                                    /* Index into the above array.
                                     */
unsigned int maski;                                  
                                    
                                    /* Size of array we load a line of the input
                                     * file into for analysis.
                                     */
#define ARRSIZ 200                  

                                    /* Small array to hold bytes from input file.
                                     */
char arr[ARRSIZ + 1];       
    
                                    /* Byte count of how far we are into the input
                                     * file.
                                     */
int  count;         
                                    /* Pointer to arr[].
                                     */
int  p;             
                                    /* Flag to say we have reached the end of the 
                                     * file, by reading an EOF byte == 0x0A.
                                     */

int eofreached = 0;                                  
                                    /* Flag to break out of file reading loop when
                                     * we reach the end. 0 means don't break.
                                     */
int infiledone = 0;

                                    /* Flag set to 1 if we have found that one 
                                     * prefix is covering an IP address already
                                     * unmasked by a previous prefix.
                                     */
int overlap = 0;

                                    /* New variables for -multi version.
                                     *
                                     * Maximum lengths of input and output
                                     * files.
                                     */
#define INFILEMAXLEN 65536
#define OUTFILEMAXLEN 65536

                                    /* Array of structures to hold data for
                                     * output file.  Each structure holds
                                     * results for the prefix which covers
                                     * the address space the ping was sent
                                     * to.
                                     *
                                     * The 'res' tag is not really needed.
                                     *
                                     * percentage = 100 * ackcount / pingsent
                                     * for the pings sent to this space.
                                     *
                                     * Likewise, count of acknowledgements and
                                     * pings sent.
                                     *
                                     * Also, the input line, including the prefix
                                     * as we read it from the input file.
                                     *
                                     * The 0th entry in this is not used.  We 
                                     * start with location 1, since that is the
                                     * lowest number line number we use.
                                     *
                                     * For instance, entry 3 in multi[] is for
                                     * the third line we selected in the input
                                     * file (after using span and offset) and
                                     * for this line, we wrote "3" into the 
                                     * mask[] integers.
                                     * 
                                     * When we do a ping to a random address, 
                                     * and this selects a mask[] number which is
                                     * non-zero, that number is the index into
                                     * multi[] where we will accumulate the
                                     * pingsent and ackcount results.
                                     *
                                     * The linenumber_m is the line number of
                                     * the original file where we found this
                                     * prefix.  
                                     */
struct res 
    {
        float percentage_m;
        int ackcount_m;
        int pingsent_m;
        int linenumber_m;
        char inputline_m[ARRSIZ+1]; 
    } multi[OUTFILEMAXLEN + 1];

                                    /* Integer to index into the above.
                                     */
int multix;
                                    /* Two new input parameters.
                                     * span's minimum value is 1.
                                     */
int span = 1;
int offset = 0;
                                    /* Counter for input lines, starting with 1.
                                     */
int linecount = 1;

                                    /* The number of prefixes we actually
                                     * selected from the input file, according to
                                     * offset and span.  This is the number of
                                     * values we find in mask[] and so it is the
                                     * number of lines we create in the main part
                                     * of the output file.
                                     *
                                     */
int resultslines = 1;
                                    /* Highest entry in multi[];
                                     *
                                     * Will be set to 1 or above for 1 or more
                                     * prefixes used.
                                     */
int multitop = 0;

                                    /* Loop counter when writing output file.
                                     */
int outloop;
                                    /* For a printing hack to provide leading
                                     * zeroes for a floating point percentage
                                     * at the start of each line in the output
                                     * file, so we can sort the lines nicely
                                     * and get them in percentage order.
                                     */
int fphacki1, fphacki2;



/* Function Prototypes
 *-------------------------------------------------------------------------
 */

void process(void);


main (int argc, char *argv[])
{
                                    /* Check input args and complain if not
                                     * the right number.
                                     */  


    if (   (argc < 3)
         ||(argc > 5)
       )  
    {
        printf("Error:\n");
        printf("There must be at least two arguments.                                        \n");
        printf(" ping-scanner-mask-multi(-nt) ffff nnnnn                                     \n");
        printf("                                                                             \n");
        printf("   ffff is the name of the input file with the masking prefixes.             \n");
        printf("                                                                             \n");
        printf("   nnnnn is 1 to 86400 for the number of seconds to ping for, typically at   \n");
        printf("   1 per second.                                                             \n");
        printf("                                                                             \n");
        printf("Optional third and fourth parameters:                                        \n");
        printf("                                                                             \n");
        printf(" ping-scanner-mask-multi(-nt) ffff nnnnn sss ooo                             \n");
        printf("                                                                             \n");
        printf("   sss  Span number to add to each line number to find the next one to read. \n");
        printf("        eg. 8 will read lines 1, 9, 17 etc.                                  \n");
        printf("                                                                             \n");
        printf("   ooo  Offset for where to start reading. 0 means no change, read from the  \n");
        printf("        first line (line number 1).  2 means read from line 3, so \"8 2\"       \n");
        printf("        means read lines 3, 11, 19 etc.                                      \n");
        printf("                                                                             \n");
        printf(" ping-scanner-mask-multi writes total results to extra line in output file   \n");
        printf(" ping-scanner-mask-multi-nt does not                                         \n");
        
        exit(1);
    }

                                    /* Set the seed from current time.
                                     */
    gettimeofday(&timenow, NULL);                                    
    randseed = timenow.tv_sec;              
    printf("randseed = %d\n", randseed);
    seedi(randseed);

                                    /* Process the command line arguments.
                                     *
                                     * "r" means open for reading text file.
                                     */
                                     
    if ( ( infile = fopen(argv[1],"r") ) == NULL)
    {
        printf("Can't open %s input file.\n",argv[1]);
        exit(1);
    }                                    
                                    /* Handle run time, span and offset 
                                     * input parameters.
                                     */ 
                                         
    runtime = atoi(argv[2]);                                         
    
    if (   (runtime < 1)
         ||(runtime > 86400)
       )
    {
        printf("\nnnnn out of range = %d.\n", runtime);
        exit(1);
    }   

                                    

    if (   (argc == 4)
         ||(argc == 5)
       )  
    {
        span = atoi(argv[3]);                                        
    
        if (   (span < 1)
             ||(span > 65536)
           )
        {
            printf("\nspan out of range = %d.\n", span);
            exit(1);
        }   
    }

    if  (argc == 5)
    {
        offset = atoi(argv[4]);                                      
    
        if (   (offset < 0)
             ||(offset >= 65536)
           )
        {
            printf("\noffset out of range = %d.\n", offset);
            exit(1);
        }   
    }

                                    /* Set up ending time.
                                     */
    starttime = timenow.tv_sec;     
    endtime = starttime + runtime;      
    

                                    /* Open output file.
                                     */
                                     
    sprintf(outfilename, "%s-results-S%d-O%d.txt", argv[1], span, offset);
    if ( ( outfile = fopen(outfilename,"w") ) == NULL)
    {
        printf("Can't open %s output file\n",outfile);
        exit(1);
    }
    
    
                                    /* Clear the input mask.
                                     */

    
    for(loopcount = 0; loopcount < MASKLEN; loopcount++)
    {
        mask[loopcount] = 0;
    }
    
    
    printf("\nOutput filename is %s.  Starting to read input file.\n",  &outfilename);


                    
    while(!feof(infile))
    {
                                        /* Read a line at a time of the input file, 
                                         * into arr[], which we soon feed to the
                                         * process() function to parse.
                                         */
        if(fgets(arr, ARRSIZ, infile))
        {

                                        /* Limit number of lines we read.
                                         */
                                         
            if (linecount > INFILEMAXLEN)
            {
                printf("Too many lines in input file!  Exiting.\n",outfile);
                exit(1);
            }
                                        /* Check that our span and offset values 
                                         * allow this line to be used.  The first
                                         * line of the file is line 1.
                                         * 
                                         * Get current line number, subtract 1,
                                         * subtract offset and then see if 
                                         * dividing it by span (modulus operator)
                                         * results in zero. 
                                         */
                                         
            if (   (((linecount - 1 - offset) % span ) == 0)
                 &&( (linecount - 1 - offset) >= 0)
               )
            {
                                        /* Process the input file to set the mask.
                                         */
                process();  
            }
            
            linecount++;

        }
    }
                                        /* Sanity check - did we process any lines?
                                         */
    if (multitop < 1)
    {
        
        printf("\nNo prefixes read.  Either no valid lines in input file, or the combination of\n");
        printf("the number of lines in the input file, and the span and offset parameters resulted\n");
        printf("in no prefixes being read.  Exiting!\n\n");
        exit(1);    
    }

    
    
    printf("Commence pinging!\n");


                                        /* The mask has been set.  Now proceed
                                         * to ping as long as there is time.
                                         */
    
    while(timenow.tv_sec < endtime) 
    {
                                                    
                                        /* Get the current time, ready for next 
                                         * while test.
                                         */
        gettimeofday(&timenow, NULL);                                    
    
                                        /* Get new pseudo-random value.
                                         */
        rawrand = ranlui(); 
        
        a2 = (rawrand & 0x00ff0000) >> 16;
        a1 = (rawrand & 0x0000ff00) >>  8;
        a0 = (rawrand & 0x000000ff);    


                                        /* Inner loop from 0 to 223, stopping 
                                         * before the multicast and reserved 
                                         * /8 prefixes.
                                         * 
                                         * Check the time too.
                                         */
        
        for(    dddcount = 0; 
                (  (dddcount < 224)
                 &&(timenow.tv_sec < endtime)
                ); 
                dddcount++
           )       
        {
                                        /* The four bytes of address are ready
                                         * to be used.  First, we need to 
                                         * check the mask to see if we are
                                         * pinging this particular /24 subnet.
                                         */
            a3 = dddcount;
            
            maski = (a3 << 16) + (a2 << 8) + a1;

                                        /* Debug.
                                         */
            
            /*  printf("test ping -w 1 %d.%d.%d.%d\n", a3, a2, a1, a0 );
             */
            
            if (mask[maski] != 0)
            {
    
                pingsent++;
                                        /* Run the ping program and check its
                                         * return status.  0 = ack received.
                                         */     
                
                                        /* First, set up multix to point to the
                                         * structure which holds results for 
                                         * whichever prefix set this mask[] integer
                                         * to non-zero.
                                         */
                                         
                multix = mask[maski];
                
                                        /* Debug.
                                         */
                /*                                       
                 * printf("multix = %d\n", multix);                                      
                 */ 
                
                                        /* Maintain ping counter for this prefix in
                                         * multi[multix].
                                         */
                                         
                multi[multix].pingsent_m++;                  
                                                         
                                         
                sprintf(commandline, "ping -w 1 %d.%d.%d.%d\n", a3, a2, a1, a0 );

                if (system(commandline) == 0)
                {
                    printf("  ------>  Ack received from %d.%d.%d.%d\n", a3, a2, a1, a0);                   
                    ackcount++;
                    
                                        /* Accumulate results into multi[multix].
                                         */
                    multi[multix].ackcount_m++;                  
                }

                                        /* Display results for this prefix only, and
                                         * then for the total results for all 
                                         * prefixes.
                                         */ 
                                         
                multi[multix].percentage_m = 100 * (float)multi[multix].ackcount_m / (float)multi[multix].pingsent_m;       

                printf("        Input file %s  Line %04d Pings sent = %05d  Acks received = %05d  Percentage = %7.4f\n", 
                                argv[1], 
                                multi[multix].linenumber_m,
                                multi[multix].pingsent_m, 
                                multi[multix].ackcount_m, 
                                multi[multix].percentage_m);    


                                        /* Display ordinary totals, as per 
                                         * ping-scanner-mask.c.
                                         */

                percentage = 100 * (float)ackcount / (float)pingsent;       
                printf("        Input file %s  Total     Pings sent = %05d  Acks received = %05d  Percentage = %7.4f\n", 
                                argv[1], 
                                pingsent, 
                                ackcount, 
                                percentage);    
            }
        }
    }
    
                                
                                    /* All pings have now been sent.
                                     *
                                     * Generate the output file lines, one for
                                     * each prefix we have been pinging.
                                     */

    for (outloop = 1; outloop < resultslines; outloop++)
    {
        
                                    /* The results line, one for each prefix
                                     * we are pinging.  The input line has a
                                     * newline at the end, so we don't need
                                     * one in the fprintf() line.
                                     *
                                     * There is a hack here.  I want a floating
                                     * point percentage value with up to three
                                     * leading zeroes to the left of the decimal
                                     * point.  AFAIK, printf() can't do this.
                                     * So I do a 3 leading zeroes integer for the
                                     * 3 digits to the left of the decimal point
                                     * and then do the right part as an integer
                                     * too.
                                     * 
                                     * Get fractional part into a floating point
                                     * variable. We multiply it by 10000 to create 
                                     * the integer to print at the right of the 
                                     * decimal point.
                                     *
                                     * I had trouble eith modf() - couldn't find
                                     * the right pointer type to keep GCC happy.
                                     * I also had trouble with fmod() - the linker
                                     * couldn't find it . . . Likewise floor()
                                     * but we don't need any of them . . 
                                     */
                                     
        fphacki1 = multi[outloop].percentage_m;
        fphacki2 = 10000 * (multi[outloop].percentage_m - fphacki1);

                                        
        fprintf(outfile, "\%03d.%04d, %05d, %05d, %05d, %s", 
                         fphacki1, fphacki2,
                         multi[outloop].ackcount_m,
                         multi[outloop].pingsent_m,
                         multi[outloop].linenumber_m,
                         multi[outloop].inputline_m
       );                               
    }
                                     
                                     
                                    /* Unless this has been commented out, as we do
                                     * for ping-scanner-mask-multi-nt.c
                                     *
                                     * Print results:  input file name, number
                                     * of pings sent, number of acks we got and the
                                     * percentage this represents.
                                     */ 
        

/*  fprintf(outfile, "\n%s,   %05d,  %05d,  %8.4f\n", argv[1], pingsent, ackcount, percentage);                             
 */ 
    return 0;
}





/*-------------------------------------------------------------------------
 *
 * Processing code - this is what does the interesting work.
 *
 * For this program, we are looking for line starts with particular numbers
 * and will spit the dummy if anything is wrong.
 *
 * Then we use the numbers to set bytes in the mask.
 *
 * Find the characters we want from the first byte of the file and
 * subsequently from the start of each line. Complain if they are
 * not what we expect.  Skip over other stuff until newline.
 * Set p to tell refill() how many characters we used and therefore
 * how many more to shift into arr[] ready for next time.
 *
 * So if the first line is 15 chars long, including the newline,
 * then arr[14] will be the newline. Set p = 14 so the whole line
 * will be pushed out of the arr[] ready for next time.
 *
 * We are looking for:
 *
 *   ddd.ccc.bbb.0/mm xxxxxxxxx(\n) 
 *
 * If the line length is 1 and the only character is a newline, then
 * we have hit the end of file.
 */

void process(void)
{
                                        /* Temp string.
                                         */
    char temp[ARRSIZ + 1];
    
    
                                        /* Pointer to within arr[].
                                         */
    char *arrp;
    
    
                                        /* Pointers to within temp[] where we
                                         * find the start of the four parts
                                         * of the prefix, and the prefix length.
                                         *
                                         * The last one, tempspace, is where we
                                         * find the space after the prefix length.
                                         */
                                         
    char *tempddd, *tempccc, *tempbbb, *tempaaa, *tempnn, *tempspace ;

                                        /* Integers for lengths of the above fields.
                                         */
    int lddd, lccc, lbbb, laaa, lnn;

                                        /* Integers for the four bytes of 
                                         * the prefix and the length.
                                         *
                                         * We expect the last byte to be 0,
                                         * so maybe we don't really need
                                         * paaa.
                                         */
    int pfddd, pfccc, pfbbb, pfaaa, pfnn;                                        

                                        /* Mask set loop counter.
                                         */
    int masksetloop;
                                        /* For calculating the span of
                                         * the bytes in the mask to set.
                                         */
                                         
    int pfspan, pfspanloop;
    
                                        /* Look for a newline within 199
                                         * characters of start of line.
                                         */
    arrp = strchr(arr, '\n');                                           

                                        /* Set p ready for refill, and exit
                                         * if the line was too long.
                                         *
                                         * Actually, p was for an old system
                                         * I am not using here . . . 
                                         */
                                        
    p = arrp - arr + 1  ;   
    if (p > 199)
    {
        printf("Line length limit of 199 exceeded. Exiting.\n");    
        exit(1);
    }
    

    if (p == 1)
    {
        printf("Last line reached.  Last prefix ddd=%d  ccc=%d  bbb=%d  nn=%d\n", pfddd, pfccc, pfbbb, pfnn);
    }
    else
    {

    
                                        /* Copy line into temp[] where we can 
                                         * work on it without upsetting arr[].
                                         *
                                         * strncpy() does not put a 0 at the end.
                                         */
        strncpy(temp, arr, p);
        temp[p] =0;



                                        /* Debugging lines
                                         */ 
/*                                                      
 *      printf("Line length = %d\n", p);        
 *      printf("Line is: >>%s<<\n", temp);      
 */ 
                                        /* -----------------------------------
                                         * 
                                         * Look for first '.'.
                                         *
                                         * Set the following:
                                         *
                                         * tempddd points to the start of the
                                         * ddd digits.
                                         *
                                         * lddd has their length.
                                         *
                                         * tempccc points to where we expect
                                         * the ccc digits to start.
                                         *
                                         * eg.
                                         *
                                         *    0123456
                                         *    12.34.56
                                         *
                                         *   lddd = 2
                                         *   tempccc = temp + 3 
                                         */

        tempddd = temp;
        lddd = (strchr(tempddd, '.') - tempddd);
        tempccc = tempddd + lddd + 1;                                   

        if (   (lddd < 1)
            ||(lddd > 3)
           )
        {
            printf("Expecting first dot at location 1 to 3 in line. Exiting.\n");   
            exit(1);
        }

                                        /* -----------------------------------
                                         * 
                                         * Look for second '.'.
                                         *
                                         * Set the following:
                                         *
                                         * lccc has their length of ccc digits.
                                         *
                                         * tempbbb points to where we expect
                                         * the bbb digits to start.
                                         */

        lccc = (strchr(tempccc, '.') - tempccc);
        tempbbb = tempccc + lccc + 1;                                   

        if (   (lccc < 1)
             ||(lccc > 3)
           )
        {
            printf("Expecting second dot 2 to 4 chars after first dot. Exiting.\n");    
            exit(1);
        }

                                        /* -----------------------------------
                                         * 
                                         * Look for third '.'.
                                         *
                                         * Set the following:
                                         *
                                         * lbbb has their length of bbb digits.
                                         *
                                         * tempaaa points to where we expect
                                         * the aaa digit (should be '0') to start.
                                         */

        lbbb = (strchr(tempbbb, '.') - tempbbb);
        tempaaa = tempbbb + lbbb + 1;                                   

        if (   (lbbb < 1)
            ||(lbbb > 3)
           )
        {
            printf("Expecting third dot 2 to 4 chars after second dot. Exiting.\n");    
            exit(1);
        }

                                        /* Check there is a '0' at tempaaa and
                                         * a '/' at tempaaa + 1.
                                         *
                                         *       "0/"
                                         *
                                         * OR    "000/"
                                         */



        if (   (tempaaa[0] == '0')
             &&(tempaaa[1] == '/')
           )
        {
            tempnn = tempaaa + 2;
        }
        else if (   (tempaaa[0] == '0')
                  &&(tempaaa[1] == '0') 
                  &&(tempaaa[2] == '/')
                )
        {
            tempnn = tempaaa + 3;
        }
        else if (   (tempaaa[0] == '0')
                  &&(tempaaa[1] == '0') 
                  &&(tempaaa[2] == '0') 
                  &&(tempaaa[3] == '/')
                )
        {
            tempnn = tempaaa + 4;
        }
        else
        {   
            printf("Expecting third dot to be followed 1, 2 or 3 zeroes and then a slash. Exiting.\n"); 
            exit(1);
        }
         

                                        /* -----------------------------------
                                         * 
                                         * Look for space after prefix length.
                                         *
                                         * Set the following:
                                         *
                                         * lnn has their length of the nn digits.
                                         */

        lnn = (strchr(tempnn, ' ') - tempnn);

        if (   (lnn < 1)
             ||(lnn > 2)
           )
        {
            printf("Expecting 1 or 2 digits of prefix length. Exiting.\n"); 
            exit(1);
        }

                                        /* Now convert these digit strings into
                                         * integers.  Note this does not check
                                         * for invalid characters like v in:
                                         *
                                         * "12v.3v.v.0/v4 " 
                                         *
                                         * We are not expecting human-generated
                                         * input - just stuff from a BGP routing
                                         * table.
                                         */
        pfddd = atoi(tempddd);
        pfccc = atoi(tempccc);
        pfbbb = atoi(tempbbb);
        pfnn =  atoi(tempnn);

                                        /* Check the ranges.
                                         */ 

        if (   (pfddd < 0)
             ||(pfddd > 223)
           )
        {
            printf("Expecting ddd in 0 to 223 range. Exiting.\n");  
            exit(1);
        }

        if (   (pfccc < 0)
             ||(pfccc > 255)
           )
        {
            printf("Expecting ccc in 0 to 255 range. Exiting.\n");  
            exit(1);
        }

        if (   (pfbbb < 0)
             ||(pfbbb > 255)
           )
        {
            printf("Expecting bbb in 0 to 255 range. Exiting.\n");  
            exit(1);
        }

        if (   (pfnn < 8)
             ||(pfnn  > 24)
           )
        {
            printf("Expecting nn in 8 to 24 range. Exiting.\n");    
            exit(1);
        }

                                        /* Debugging line.
                                         */     
/*                                                  
 *      printf("Found: ddd=%d  ccc=%d  bbb=%d  nn=%d p=%d\n", pfddd, pfccc, pfbbb, pfnn, p);
 */

                                        /* Set the mask bytes.  We don't need 
                                         * checking to see if we the mask[] array 
                                         * since it has a /8 worth of spare 
                                         * space at the end.
                                         *
                                         * The prefix length tells us how many
                                         * addresses to step.  These BGP routing
                                         * table entries often overlap binary
                                         * boundaries, for instance:
                                         *  
                                         * 0000    192.126.0.0/24 256  rir
                                         * 0001  a 192.126.1.0/22 1024 adv
                                         * 0010  a
                                         * 0011  a
                                         * 0100  a
                                         * 0101    192.126.5.0/24 256  unadv
                                         * 0110  a 192.126.6.0/24 256  adv
                                         * 0111    192.126.7.0/24 256  unadv
                                         */
                                            
                                         
        maski = (pfddd << 16) + (pfccc << 8) +  pfbbb;
    
                                        /* Now to interpret these and set 
                                         * bytes in mask[] which correspond
                                         * to this prefix, to the line number
                                         * in the input file.  We also
                                         * check to see if one of these
                                         * bytes was previously set to !=0
                                         * and issue a warning to the console
                                         * for the first instance of this.
                                         *
                                         * A /24 will set a single byte.
                                         * A /23 will set 2 bytes etc.
                                         * A /8 will set 2^16 bytes.
                                         */

                
        pfspan = 1;

        for (pfspanloop = 24 - pfnn; pfspanloop > 0; pfspanloop--)
        {
            pfspan = pfspan << 1;
        }

                                        /* Set the bytes to resultslines, starting
                                         * with maski and working upwards.
                                         *
                                         * resultslines = 1 means the first prefix
                                         * we are handling. 2 means the next etc.
                                         * 
                                         * If span was 2 and offset was 3 then
                                         * we would have this relationship between
                                         * input line numbers and resultlines:
                                         *
                                         *   1
                                         *   2
                                         *   3    1
                                         *   4
                                         *   5    2
                                         *   6
                                         *   7    3
                                         */
                                         
        for (pfspanloop =0; pfspanloop < pfspan; pfspanloop++)
        {
            if (   (overlap == 0)
                 &&(mask[maski + pfspanloop] != 0)
               )
            {
                overlap = 1;
                printf("\nPrefix %d.%d.%d.0  caused overlap %d /24s beyond start! Any subsequent occurences not reported.\n", pfddd, pfccc, pfbbb, pfspanloop); 
            }
            
            mask[maski + pfspanloop] = resultslines;
            
                                        /* Debug.
                                         */
             printf("Setting %d to %d\n", maski + pfspanloop, resultslines);
            
        }
 
                                        /* Write various things into multi[].
                                         */

        multi[resultslines].percentage_m = 0;
        multi[resultslines].ackcount_m = 0;
        multi[resultslines].pingsent_m = 0;
        multi[resultslines].linenumber_m = linecount;
        
        
        sprintf(multi[resultslines].inputline_m, "%s", temp);
                                          
                                        /* Write to the integer where we keep the
                                         * highest entry in multi[] so far.
                                         */
        multitop = resultslines;                                           
                                          
                                        /* Increment resultslines so next time process()
                                         * is called, it will write into the next res
                                         * structure in multi[].
                                         */
        resultslines++;
    }
}


