/* ping-scanner-mask.c          Version 1.00  2007 February 23
 *
 * This program pings random IPv4 addresses according to a mask set by reading
 * a text file of one or more prefixes.
 * 
 * Pings will be sent randomly within the allowed time, usually once a second, for 
 * the number of pings specified.  However, since the timing is done by the
 * external ping program, sometimes the pings may be sent faster than this.
 *
 * An example of the format of the input file is:
 *
 * 192.231.221.0/24 256 19921022 arin adv
 * 192.231.233.0/24 256 19921022 arin adv
 * 192.231.253.0/24 256 19921026 arin adv   
 *
 * where all that matters is the lines start with:
 *
 * "ddd.ccc.bbb.0/mm "  
 *
 * where ddd, ccc and aaa are 1 to 3 digits 0 to 255 and where mm is 1 or 2
 * digits 1 to 24.  The maximum line length is 200 chars.
 * 
 * There must be a space, and then there can be 0 or more other characters.
 *
 * It is OK for the prefix lengths and addresses not to be on binary boundaries
 * so:
 *
 *   1.2.3.0/24 
 *   1.2.3.1/23
 *
 * would cover:
 *  
 *   1.2.3.0
 *   1.2.3.1
 *   1.2.3.2
 *
 * If one prefix overlaps another, a single error will be reported on the 
 * console for the first occurence, but the program will continue. 
 *
 * There will be repetitive pinging of those smaller masks where the number 
 * of available IP addresses is less than the total number of pings.
 *
 * The 31 bit PSRNG is seeded on the current time, so results will differ each 
 * time.  Bits 0 to 23 of the PSRNG are used, for these bits of the address.
 * Bits 24 to 31 of the address are done with a counter.
 * 
 * So for a given random number ccc.bbb.aaa, the program will potentially ping
 * all addressses 0.ccc.bbb.aaa to 223.ccc.bbb.aaa, according to the mask.
 *
 * The program counts the number of responses, which arrive in less than a second, 
 * and writes the total to a text file with the same name as the input file, 
 * with "-results.txt" appended to its name.  It also writes the number of pings
 * sent and number of acknowledgements as a  percentage of the number of pings sent.
 *
 * The purpose is to try to estimate the percentage of IP address space which actually has a
 * computer connected to it, for a particular subset of the address space.  For instance
 * by grepping http://bgp.potaroo.net/ipv4-stats/prefixes.txt for all the prefixes of a
 * particular length.
 *
 * Compile with the GCC C compiler:
 * 
 *   gcc ping-scanner-mask.c -o ping-scanner-mask
 *
 * Check the -w or -t option of your ping program - the one which enables it to exit
 * after a certain time.  Some are -w and some are -t.  Below, I assume -w.
 *
 * The ping program may return in less than a second if it gets an ack packet (including
 * an ack from another IP address, such a router en-route which takes it upon itself
 * to acknowlede pings to a now non-existent subnet).  In order to keep the process 
 * going at a reliable rate, such as for a 24 hour survey, it might be possible to 
 * code some timing functions into this program, but I have chosen not to, since
 * I think that would tie up the CPU waiting for a second to elapse, and that this
 * would get in the way of using multiple instances of this program in parallel, due
 * to CPU load.  For instance, we might want to run an instance for every prefix length
 * /8 to /24.
 *
 * Copyright public domain - Robin Whittle hhtp://www.firstpr.com.au/ip/
 *
 * Please use this judiciously.  Folks may be upset about being pinged, 
 * especially if they are the US Department of Defence!
 *
 */



#include <stdio.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include "rand31-park-miller-carta.c"

                                    /* Global variables.
                                     */
                                     
long unsigned int loopcount;
                                    /* Bytes 3, 2, 1 and 0 of address to ping.
                                     */
unsigned int a3, a2, a1, a0;
    
                                    /* Number of pings sent so far.
                                     */
long int pingsent = 0;

                                    /* Number of acknowledgements received so far.
                                     */
long int ackcount = 0;
                                    /* Percentage of pings returned.
                                     */
float percentage = 0;    
                                    /* Raw random number.
                                     */ 
int rawrand;
    
                                    /* Counter for ddd byte cycle.
                                     */
int dddcount =0;                                                                        
                                    
                                    /* String to prepare the command line in.
                                     */
char commandline[80];
    
                                    /* String to prepare the output file
                                     * name in.
                                     */
char outfilename[80];

                                    /* Input and output file buffers.
                                     */ 
FILE *infile;                                                        
FILE *outfile;                                                       


                                    /* Seed for random number generator.
                                     */
long unsigned int randseed = 0;
    
                                    /* Structure for gettimeofday().
                                     */
struct timeval timenow;
                                    /* Absolute time in seconds when the program
                                     * starts, when it should stop and how many 
                                     * seconds it should run for.
                                     */
                                      
int starttime, endtime, runtime; 


                                    /* 2^24 bytes, each of which is set to
                                     * 0 unless this particular /24 is enabled
                                     * for pinging due to being included in
                                     * one of more of the prefixes we read in
                                     * the input file.
                                     */
    
char mask[16777216];

                                    /* Index into the above array.
                                     */
unsigned int maski;                                  
                                    
                                    /* Size of array we load a line of the input
                                     * file into for analysis.
                                     */
#define ARRSIZ 200                  

                                    /* Small array to hold bytes from input file.
                                     */
char arr[ARRSIZ + 1];       
    
                                    /* Byte count of how far we are into the input
                                     * file.
                                     */
int  count;         
                                    /* Pointer to arr[]
                                     */
int  p;             
                                    /* Flag to say we have reached the end of the 
                                     * file, by reading an EOF byte == 0x0A.
                                     */

int eofreached = 0;                                  
                                    /* Flag to break out of file reading loop when
                                     * we reach the end. 0 means don't break.
                                     */
int infiledone = 0;

                                    /* Flag set to 1 if we have found that one 
                                     * prefix is covering an IP address already
                                     * unmasked by a previous prefix.
                                     */
int overlap = 0;


/* Function Prototypes
 *-------------------------------------------------------------------------
 */

void process(void);


main (int argc, char *argv[])
{
                                    /* Check input args and complain if not
                                     * the right number.
                                     */  


    if (argc < 3)
    {
        printf("Error:\n");
        printf("There must be at least two arguments.\n");
        printf(" ping-scanner-mask ffff nnnnn\n");
        printf("                                             \n");
        printf("   ffff is the name of the input file with the masking prefixes.\n");
        printf("   nnnnn is 1 to 86400 for the number of seconds to ping for, typically at 1 per second.\n\n");
        exit(1);
    }

                                    /* Set the seed from current time.
                                     */
    gettimeofday(&timenow, NULL);                                    
    randseed = timenow.tv_sec;              
    printf("randseed = %d\n", randseed);
    seedi(randseed);

                                    /* Process the command line arguments.
                                     *
                                     * "r" means open for reading text file.
                                     */
                                     
    if ( ( infile = fopen(argv[1],"r") ) == NULL)
    {
        printf("Can't open %s input file.\n",argv[1]);
        exit(1);
    }                                    
                                     
                                         
    runtime = atoi(argv[2]);                                         
    
    if (   (runtime < 1)
         ||(runtime > 86400)
       )
    {
        printf("\nnnnn out of range = %d.\n", runtime);
        exit(1);
    }   

                                    /* Set up ending time.
                                     */
    starttime = timenow.tv_sec;     
    endtime = starttime + runtime;      
    

                                    /* Open output file.
                                     */
    sprintf(outfilename, "%s-results.txt", argv[1]);
    if ( ( outfile = fopen(outfilename,"w") ) == NULL)
    {
        printf("Can't open %s output file\n",outfile);
        exit(1);
    }
    
    
                                    /* Clear the input mask.
                                     */

    
    for(loopcount = 0; loopcount < 16777216; loopcount++)
    {
        mask[loopcount] = 0;
    }
    
    
    printf("\nOutput filename is %s.  Starting to read input file.\n",  &outfilename);


                    
    while(!feof(infile))
    {
        if(fgets(arr, ARRSIZ, infile))
        {
                                        /* Process the input file to set the mask.
                                         */
            process();  
        }
    }
    
    printf("Commence pinging!\n");


                                        /* The mask has been set.  Now proceed
                                         * to ping as long as there is time.
                                         */
    
    while(timenow.tv_sec < endtime) 
    {
                                                    
                                        /* Get the current time, ready for next 
                                         * while test.
                                         */
        gettimeofday(&timenow, NULL);                                    
    
                                        /* Get new pseudo-random value.
                                         */
        rawrand = ranlui(); 
        
        a2 = (rawrand & 0x00ff0000) >> 16;
        a1 = (rawrand & 0x0000ff00) >>  8;
        a0 = (rawrand & 0x000000ff);    


                                        /* Inner loop from 0 to 223, stopping 
                                         * before the multicast and reserved 
                                         * /8 prefixes.
                                         * 
                                         * Check the time too.
                                         */
        
        for(    dddcount = 0; 
                (  (dddcount < 224)
                 &&(timenow.tv_sec < endtime)
                ); 
                dddcount++
           )       
        {
                                        /* The four bytes of address are ready
                                         * to be used.  First, we need to 
                                         * check the mask to see if we are
                                         * pinging this particular /24 subnet.
                                         */
            a3 = dddcount;
            
            maski = (a3 << 16) + (a2 << 8) + a1;

                                        /* Debug.
                                         */
            
            /*  printf("test ping -w 1 %d.%d.%d.%d\n", a3, a2, a1, a0 );
             */
            
            if (mask[maski] != 0)
            {
    
                pingsent++;
                                        /* Run the ping program and check its
                                         * return status.  0 = ack received.
                                         */     
                                         
                sprintf(commandline, "ping -w 1 %d.%d.%d.%d\n", a3, a2, a1, a0 );

                if (system(commandline) == 0)
                {
                    printf("  ------>  Ack received from %d.%d.%d.%d\n", a3, a2, a1, a0);                   
                    ackcount++;
                }

                percentage = 100 * (float)ackcount / (float)pingsent;       
                printf("             Input file %s        Pings sent = %05d  Acks received = %05d  Percentage = %8.4f\n", argv[1], pingsent, ackcount, percentage); 

            }
        }
    }
    
                                
                                    /* All pings have now been sent.
                                     *
                                     * Print results:  input file name, number
                                     * of pings sent, number of acks we got and the
                                     * percentage this represents.
                                     */ 
        

    fprintf(outfile, "%s,   %05d,  %05d,  %8.4f\n", argv[1], pingsent, ackcount, percentage);                               
    
    return 0;
}





/*-------------------------------------------------------------------------
 *
 * Processing code - this is what does the interesting work.
 *
 * For this program, we are looking for line starts with particular numbers
 * and will spit the dummy if anything is wrong.
 *
 * Then we use the numbers to set bytes in the mask.
 *
 * Find the characters we want from the first byte of the file and
 * subsequently from the start of each line. Complain if they are
 * not what we expect.  Skip over other stuff until newline.
 * Set p to tell refill() how many characters we used and therefore
 * how many more to shift into arr[] ready for next time.
 *
 * So if the first line is 15 chars long, including the newline,
 * then arr[14] will be the newline. Set p = 14 so the whole line
 * will be pushed out of the arr[] ready for next time.
 *
 * We are looking for:
 *
 *   ddd.ccc.bbb.0/mm xxxxxxxxx(\n) 
 *
 * If the line length is 1 and the only character is a newline, then
 * we have hit the end of file.
 */

void process(void)
{
                                        /* Temp string.
                                         */
    char temp[ARRSIZ + 1];
    
    
                                        /* Short string for 1 to 3 decimal digits 
                                         * for conversion to numbers.
                                         */
    char digits[80];
    
    
                                        /* Pointer to within arr[].
                                         */
    char *arrp;
    
    
                                        /* Pointers to within temp[] where we
                                         * find the start of the four parts
                                         * of the prefix, and the prefix length.
                                         *
                                         * The last one, tempspace, is where we
                                         * find the space after the prefix length.
                                         */
                                         
    char *tempddd, *tempccc, *tempbbb, *tempaaa, *tempnn, *tempspace ;

                                        /* Integers for lengths of the above fields.
                                         */
    int lddd, lccc, lbbb, laaa, lnn;

                                        /* Integers for the four bytes of 
                                         * the prefix and the length.
                                         *
                                         * We expect the last byte to be 0,
                                         * so maybe we don't really need
                                         * paaa.
                                         */
    int pfddd, pfccc, pfbbb, pfaaa, pfnn;                                        

                                        /* Mask set loop counter.
                                         */
    int masksetloop;
                                        /* For calculating the span of
                                         * the bytes in the mask to set.
                                         */
                                         
    int pfspan, pfspanloop;
    
                                        /* Look for a newline within 199
                                         * characters of start of line.
                                         */
    arrp = strchr(arr, '\n');                                           

                                        /* Set p ready for refill, and exit
                                         * if the line was too long.
                                         */
                                        
    p = arrp - arr + 1  ;   
    if (p > 199)
    {
        printf("Line length limit of 199 exceeded. Exiting.\n");    
        exit(1);
    }
    

    if (p == 1)
    {
        printf("Last line reached.  Last prefix ddd=%d  ccc=%d  bbb=%d  nn=%d\n", pfddd, pfccc, pfbbb, pfnn);
    }
    else
    {

    
                                        /* Copy line into temp[] where we can 
                                         * work on it without upsetting arr[].
                                         *
                                         * strncpy() does not put a 0 at the end.
                                         */
        strncpy(temp, arr, p);
        temp[p] =0;



                                        /* Debugging lines
                                         */ 
/*                                                      
 *      printf("Line length = %d\n", p);        
 *      printf("Line is: >>%s<<\n", temp);      
 */ 
                                        /* -----------------------------------
                                         * 
                                         * Look for first '.'.
                                         *
                                         * Set the following:
                                         *
                                         * tempddd points to the start of the
                                         * ddd digits.
                                         *
                                         * lddd has their length.
                                         *
                                         * tempccc points to where we expect
                                         * the ccc digits to start.
                                         *
                                         * eg.
                                         *
                                         *    0123456
                                         *    12.34.56
                                         *
                                         *   lddd = 2
                                         *   tempccc = temp + 3 
                                         */

        tempddd = temp;
        lddd = (strchr(tempddd, '.') - tempddd);
        tempccc = tempddd + lddd + 1;                                   

        if (   (lddd < 1)
            ||(lddd > 3)
           )
        {
            printf("Expecting first dot at location 1 to 3 in line. Exiting.\n");   
            exit(1);
        }

                                        /* -----------------------------------
                                         * 
                                         * Look for second '.'.
                                         *
                                         * Set the following:
                                         *
                                         * lccc has their length of ccc digits.
                                         *
                                         * tempbbb points to where we expect
                                         * the bbb digits to start.
                                         */

        lccc = (strchr(tempccc, '.') - tempccc);
        tempbbb = tempccc + lccc + 1;                                   

        if (   (lccc < 1)
             ||(lccc > 3)
           )
        {
            printf("Expecting second dot 2 to 4 chars after first dot. Exiting.\n");    
            exit(1);
        }

                                        /* -----------------------------------
                                         * 
                                         * Look for third '.'.
                                         *
                                         * Set the following:
                                         *
                                         * lbbb has their length of bbb digits.
                                         *
                                         * tempaaa points to where we expect
                                         * the aaa digit (should be '0') to start.
                                         */

        lbbb = (strchr(tempbbb, '.') - tempbbb);
        tempaaa = tempbbb + lbbb + 1;                                   

        if (   (lbbb < 1)
            ||(lbbb > 3)
           )
        {
            printf("Expecting third dot 2 to 4 chars after second dot. Exiting.\n");    
            exit(1);
        }

                                        /* Check there is a '0' at tempaaa and
                                         * a '/' at tempaaa + 1
                                         */

        if (   (tempaaa[0] != '0')
             ||(tempaaa[1] != '/')
           )
        {
            printf("Expecting third dot to be followed by zero and then a slash. Exiting.\n");  
            exit(1);
        }
         

                                        /* -----------------------------------
                                         * 
                                         * Look for space after prefix length.
                                         *
                                         * Set the following:
                                         *
                                         * lnn has their length of the nn digits.
                                         */

        tempnn = tempaaa + 2;
        lnn = (strchr(tempnn, ' ') - tempnn);

        if (   (lnn < 1)
             ||(lnn > 2)
           )
        {
            printf("Expecting 1 or 2 digits of prefix length. Exiting.\n"); 
            exit(1);
        }

                                        /* Now convert these digit strings into
                                         * integers.  Note this does not check
                                         * for invalid characters like v in:
                                         *
                                         * "12v.3v.v.0/v4 " 
                                         *
                                         * We are not expecting human-generated
                                         * input - just stuff from a BGP routing
                                         * table.
                                         */
        pfddd = atoi(tempddd);
        pfccc = atoi(tempccc);
        pfbbb = atoi(tempbbb);
        pfnn =  atoi(tempnn);

                                        /* Check the ranges.
                                         */ 

        if (   (pfddd < 0)
             ||(pfddd > 223)
           )
        {
            printf("Expecting ddd in 0 to 223 range. Exiting.\n");  
            exit(1);
        }

        if (   (pfccc < 0)
             ||(pfccc > 255)
           )
        {
            printf("Expecting ccc in 0 to 255 range. Exiting.\n");  
            exit(1);
        }

        if (   (pfbbb < 0)
             ||(pfbbb > 255)
           )
        {
            printf("Expecting bbb in 0 to 255 range. Exiting.\n");  
            exit(1);
        }

        if (   (pfnn < 8)
             ||(pfnn  > 24)
           )
        {
            printf("Expecting nn in 8 to 24 range. Exiting.\n");    
            exit(1);
        }

                                        /* Debugging line.
                                         */     
/*                                                  
 *      printf("Found: ddd=%d  ccc=%d  bbb=%d  nn=%d p=%d\n", pfddd, pfccc, pfbbb, pfnn, p);
 */

                                        /* Set the mask bytes.  We don't need to
                                         * checking to see if we the mask[] array 
                                         * since it has multicast and reserved
                                         * stuff at the end anyway.
                                         *
                                         * The prefix length tells us how many
                                         * addresses to step.  These BGP routing
                                         * table entries often overlap binary
                                         * boundaries, for instance:
                                         *  
                                         * 0000    192.126.0.0/24 256  rir
                                         * 0001  a 192.126.1.0/22 1024 adv
                                         * 0010  a
                                         * 0011  a
                                         * 0100  a
                                         * 0101    192.126.5.0/24 256  unadv
                                         * 0110  a 192.126.6.0/24 256  adv
                                         * 0111    192.126.7.0/24 256  unadv
                                         */
                                            
                                         
        maski = (pfddd << 16) + (pfccc << 8) +  pfbbb;
    
                                        /* Now to interpret these and set 
                                         * bytes in mask[] which correspond
                                         * to this prefix, to 01.  We also
                                         * check to see if one of these
                                         * bytes was previously set to 01
                                         * and issue a warning to the console
                                         * for the first instance of this.
                                         *
                                         * A /24 will set a single byte.
                                         * A /23 will set 2 bytes etc.
                                         * A /8 will set 2^16 bytes.
                                         */

                
        pfspan = 1;

        for (pfspanloop = 24 - pfnn; pfspanloop > 0; pfspanloop--)
        {
            pfspan = pfspan << 1;
        }

                                        /* Set the bytes to 1, starting
                                         * with maski and working upwards.
                                         */
                                         
        for (pfspanloop =0; pfspanloop < pfspan; pfspanloop++)
        {
            if (   (overlap == 0)
                 &&(mask[maski + pfspanloop] != 0)
               )
            {
                overlap = 1;
                printf("\nPrefix %d.%d.%d.0  caused overlap %d /24s beyond start! Any subsequent occurences not reported.\n", pfddd, pfccc, pfbbb, pfspanloop); 
            }
            
            mask[maski + pfspanloop] = 1;
            
                                        /* Debug.
                                         */
            /* printf("Setting %d\n", maski + pfspanloop);
             */
        }
 
    }
}


