/* ping-scanner.c          Version 1.00  2007 February 19
 *
 * This program pings random IP addresses in a particular IPv4 prefix, such as 203.0.0.0/8.
 * It counts the number of responses, which arrive in less than a second, and writes
 * the total to a text file with the same name as the prefix.
 *
 * The command line format is in some printf()s below.  One particular /8 is specified, 
 * together with the number of pings to send.
 *
 * The idea is that a second program ping-scanner-call-multiple.c will run multiple
 * instances of this program, to simultaneously try to send 1000 or so packets to 
 * a number of contiguous prefixes at the same time.
 *
 * The purpose is to try to estimate the percentage of IP address space which actually has a
 * computer connected to it.
 *
 *
 * Compile with the GCC C compiler:
 * 
 *   gcc ping-scanner.c -o ping-scanner
 *
 * Check the -w or -t option of your ping program - the one which enables it to exit
 * after a certain time.  Some are -w and some are -t.  Below, I assume -w.
 *
 * This program should be called with a number ppp 1 to 255 to set the /8 prefix to scan
 * and the number of pings to send before exiting.  On exit, it writes to a file xxx.txt
 * the number of responses received.  It also writes to the console the current status.
 *
 * Copyright public domain - Robin Whittle hhtp://www.firstpr.com.au/ip/
 *
 * Please use this judiciously.  Folks may be upset about being pinged, 
 * especially if they are the US Department of Defence!
 *
 */

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include "rand31-park-miller-carta.c"

main (int argc, char *argv[])
{
                                    /* 32 bit unsigned integer
                                     * to store new pseudo-random numbers.
                                     */
    long unsigned int test3, test2, test1, test0;

    long unsigned int loopcount;
                                    /* Bytes 3, 2, 1 and 0 of address to ping.
                                     */
    unsigned int a3, a2, a1, a0;
    
                                    /* Number of pings returned so far.
                                     */
    
    long int pingcount = 0;
                                    /* Percentage of pings returned.
                                     */
    float percentage = 0;    
                                    /* String to prepare the command line in.
                                     */
    char commandline[80];
    
                                    /* String to prepare the output filen name in.
                                     */
    char outfilename[80];

    FILE *outfile;                  /* Output file buffer.
                                     */ 

                                    /* How many pings to send.
                                     */
    int looplength; 

                                    /* Seed for random number generator.
                                     */
    long unsigned int randseed = 0;
    
                                    /* Structure for gettimeofday().
                                     */
    struct timeval timenow;



    if (argc < 3)
    {
        printf("Error:\n");
        printf("There must be at least two numeric arguments.\n");
        printf(" ping-scanner ppp nnnnn\n");
        printf("                                             \n");
        printf("   ppp is 1 to 255 for the subnet number.\n");
        printf("   nnnnn is 1 to 86400 for the number of pings to send.\n\n");
        exit(1);
    }

                                    /* Process the command line arguments.
                                     */
                                         
    a3 =         atoi(argv[1]);                                      
    looplength = atoi(argv[2]);                                      
    
    if (   (a3 < 1)
         ||(a3 > 255)
       )
    {
        printf("\nppp out of range = %d.\n\n", a3);
        exit(1);
    }     
    
    if (   (looplength < 1)
         ||(looplength > 86400)
       )
    {
        printf("\nnnnn out of range = %d.\n", looplength);
        exit(1);
    }   


                                    /* Open output file with name such as "013.txt"
                                     */
    sprintf(outfilename, "%03d.txt", a3);
    if ( ( outfile = fopen(outfilename,"w") ) == NULL)
    {
        printf("Can't open %s output file\n",outfile);
        exit(1);
    }
    
    
    printf("\nLets Go!  Pinging prefix %d with %d attempts. Output filename is %s\n", a3, looplength, &outfilename);
    

                                    /* Set the seed from current time plus the subnet 
                                     * number.
                                     */
    gettimeofday(&timenow, NULL);                                    
    randseed = timenow.tv_sec + a3;             
    printf("randseed = %d\n", randseed);
                                     
    seedi(randseed);
    
    
    for(loopcount = 0; loopcount < (unsigned)looplength; loopcount++)
    {
                                    /* Get new values from the generator.
                                     */
        test2 = ranlui();   
        test1 = ranlui();   
        test0 = ranlui();   
    
        a2 = test2 & 255;
        a1 = test1 & 255;
        a0 = test0 & 255;   
    
        sprintf(commandline, "ping -w 1 %d.%d.%d.%d\n", a3, a2, a1, a0 );

        if (system(commandline) == 0)
        {
            pingcount++;
        }
        
        printf("Run %d, packets recieved = %d\n", loopcount, pingcount );   
            
    }
                                    /* Print results: prefix number, number
                                     * of response packets we got and the
                                     * percentage this represents.
                                     */ 
        
        percentage = 100 * (float)pingcount / (float)looplength;

        fprintf(outfile, "%03d,  %05d,  %8.4f\n", a3, pingcount, percentage);                               
    
    return 0;
}
