/* prefix-span-tag-zeropad.c          Version 1.00  2007 February 27
 *
 * Reads a text file with lines of the form:
 *
 * 192.42.99.0/24 256 19880916 afrinic adv
 *
 * and writes a new file, with the addresses and prefix
 * length zero extended and "Dxx" added:
 *
 * 192.042.099.000/24 , D00 , 256 19880916 afrinic adv  
 *
 * This will change a /8 to a /08!
 * 
 * Where D00 means this prefix does not cross over any
 * borders of longer prefixes.
 *
 * Prefixes up to 32 are allowed in the input file, but
 * are not considered when writing the output file.
 *
 *
 * In the example:
 *  
 * 0000    192.126.0.0/24 256  rir
 * 0001  a 192.126.1.0/22 1024 adv
 * 0010  a
 * 0011  a
 * 0100  a
 * 0101    192.126.5.0/24 256  unadv
 * 0110  a 192.126.6.0/24 256  adv
 * 0111    192.126.7.0/24 256  unadv
 *
 * The second prefixes crosses boundaries of a /24, so its "depth" is
 * 2. That line will be returned as:
 *
 * 192.126.001.000/22 , D02 , 1024 adv      
 *
 * This only works for prefixes between /8 and /24.
 * Lines with prefixes of longer lengths will not appear in the output file.
 *
 * Then it is easy to separate lines out with grep, into files with prefixes
 * of different lengths, advertised or not, spanning longer prefix boundaries
 * etc.
 *
 * Compiled with GCC:
 *
 *    gcc prefix-span-tag-zeropad.c -o prefix-span-tag-zeropad
 *  
 * Copyright public domain - Robin Whittle hhtp://www.firstpr.com.au/ip/
 *
 */



#include <stdio.h>
#include <string.h>

                                    /* Global variables.
                                     */
                                     
long unsigned int loopcount;
                                    /* String to prepare the output file
                                     * name in.
                                     */
char outfilename[80];

                                    /* Input and output file buffers.
                                     */ 
FILE *infile;                                                        
FILE *outfile;                                                       

                                    /* Size of array we load a line of the input
                                     * file into for analysis.
                                     */
#define ARRSIZ 200                  

                                    /* Small array to hold bytes from input file.
                                     */
char arr[ARRSIZ + 1];       
    
                                    /* Byte count of how far we are into the input
                                     * file.
                                     */
int  count;         
                                    /* Pointer to arr[]
                                     */
int  p;             
                                    /* Flag to say we have reached the end of the 
                                     * file, by reading an EOF byte == 0x0A.
                                     */

int eofreached = 0;                                  

                                    /* Set to ignore the current input line.
                                     */
int ignoreline = 0;

                                    /* Flag to break out of file reading loop when
                                     * we reach the end. 0 means don't break.
                                     */
int infiledone = 0;

                                    /* Counter for input lines, starting with 1.
                                     */
int linecount = 1;

                                    /* Loop counter when writing output file.
                                     */
int outloop;



/* Function Prototypes
 *-------------------------------------------------------------------------
 */

void process(void);


main (int argc, char *argv[])
{
                                    /* Check input args and complain if not
                                     * the right number.
                                     */  


    if (argc != 3) 
    {
        printf("Error:                                                                       \n");   
        printf("There must be at two arguments.                                              \n");
        printf(" prefix-span-tag iiii oooo                                                   \n");
        printf("                                                                             \n");
        printf("   iiii is the name of the input file.                                       \n");
        printf("   oooo is the name of the output file.                                      \n");
        
        exit(1);
    }

                                    /* Process the command line arguments.
                                     *
                                     * "r" means open for reading text file.
                                     */
                                     
    if ( ( infile = fopen(argv[1],"r") ) == NULL)
    {
        printf("Can't open %s input file.\n",argv[1]);
        exit(1);
    }                                    
                                    /* Handle run time, span and offset 
                                     * input parameters.
                                     */ 
                                    /* Open output file.
                                     */
                                     
    if ( ( outfile = fopen(argv[2],"w") ) == NULL)
    {
        printf("Can't open %s output file\n",outfile);
        exit(1);
    }

                    
    while(!feof(infile))
    {
                                        /* Read a line at a time of the input file, 
                                         * into arr[], which we soon feed to the
                                         * process() function to parse.
                                         */
        if(fgets(arr, ARRSIZ, infile))
                                        /* Process the input file and
                                         * write the output file.
                                         */
        {
            process();  
        }
    }
    
    return 0;
}





/*-------------------------------------------------------------------------
 *
 * Processing code - this is what does the interesting work.
 *
 * For this program, we are looking for line starts with particular numbers
 * and will spit the dummy if anything is wrong.
 *
 * Then we use the numbers and the rest of the line to generate the output 
 * line.
 *
 * Find the characters we want from the first byte of the file and
 * subsequently from the start of each line. Complain if they are
 * not what we expect.  Skip over other stuff until newline.
 * Set p to tell refill() how many characters we used and therefore
 * how many more to shift into arr[] ready for next time.
 *
 * So if the first line is 15 chars long, including the newline,
 * then arr[14] will be the newline. Set p = 14 so the whole line
 * will be pushed out of the arr[] ready for next time.
 *
 * We are looking for:
 *
 *   ddd.ccc.bbb.0/mm xxxxxxxxx(\n) 
 *
 * If the line length is 1 and the only character is a newline, then
 * we have hit the end of file.
 */

void process(void)
{
                                        /* Temp string.
                                         */
    char temp[ARRSIZ + 1];
    
    
                                        /* Pointer to within arr[].
                                         */
    char *arrp;
    
    
                                        /* Pointers to within temp[] where we
                                         * find the start of the four parts
                                         * of the prefix, and the prefix length.
                                         *
                                         * The last one, tempspace, is where we
                                         * find the space after the prefix length.
                                         */
                                         
    char *tempddd, *tempccc, *tempbbb, *tempaaa, *tempnn, *tempspace ;

                                        /* Integers for lengths of the above fields.
                                         */
    int lddd, lccc, lbbb, laaa, lnn;

                                        /* Integers for the four bytes of 
                                         * the prefix and the length.
                                         */
    int pfddd, pfccc, pfbbb, pfaaa, pfnn;                                        

                                        /* The address of the prefix.
                                         */
    int pfaddress;
    
                                        /* The lowest bit number of the address 
                                         * which has bit 1 set.
                                         */
    int lowestbit;
    
                                        /* How many bits beyond the current 
                                         * prefix length this prefix steps
                                         * over bit boundaries.
                                         *
                                         * = 32 - pfnn - lowestbit
                                         */
    int depth;
    
                                        /* Look for a newline within 199
                                         * characters of start of line.
                                         */
    arrp = strchr(arr, '\n');                                           

                                        /* Set p ready for refill, and exit
                                         * if the line was too long.
                                         * 
                                         * Actually, p was for an old system
                                         * I am not using here . . .
                                         */
                                        
    p = arrp - arr + 1  ;   
    if (p > 199)
    {
        printf("Line length limit of 199 exceeded. Exiting.\n");    
        exit(1);
    }
    
    if (p == 1)
    {
        printf("Last line reached.  Last prefix ddd=%d  ccc=%d  bbb=%d  nn=%d\n", pfddd, pfccc, pfbbb, pfnn);
    }
    else
    {   
                                        /* Copy line into temp[] where we can 
                                         * work on it without upsetting arr[].
                                         *
                                         * strncpy() does not put a 0 at the end.
                                         */
        strncpy(temp, arr, p);
        temp[p] =0;



                                        /* Debugging lines
                                         */ 
/*                                                      
 *      printf("Line length = %d\n", p);        
 *      printf("Line is: >>%s<<\n", temp);      
 */ 
                                        /* -----------------------------------
                                         * 
                                         * Look for first '.'.
                                         *
                                         * Set the following:
                                         *
                                         * tempddd points to the start of the
                                         * ddd digits.
                                         *
                                         * lddd has their length.
                                         *
                                         * tempccc points to where we expect
                                         * the ccc digits to start.
                                         *
                                         * eg.
                                         *
                                         *    0123456
                                         *    12.34.56
                                         *
                                         *   lddd = 2
                                         *   tempccc = temp + 3 
                                         */

        tempddd = temp;
        lddd = (strchr(tempddd, '.') - tempddd);
        tempccc = tempddd + lddd + 1;                                   

        if (   (lddd < 1)
             ||(lddd > 3)
           )
        {
            printf("Expecting first dot at location 1 to 3 in line. Exiting. p = %d string = %s\n", p, arr);    
            exit(1);
        }

                                        /* -----------------------------------
                                         * 
                                         * Look for second '.'.
                                         *
                                         * Set the following:
                                         *
                                         * lccc has the length of ccc digits.
                                         *
                                         * tempbbb points to where we expect
                                         * the bbb digits to start.
                                         */

        lccc = (strchr(tempccc, '.') - tempccc);
        tempbbb = tempccc + lccc + 1;                                   

        if (   (lccc < 1)
             ||(lccc > 3)
           )
        {
            printf("Expecting second dot 2 to 4 chars after first dot. Exiting.\n");    
            exit(1);
        }

                                        /* -----------------------------------
                                         * 
                                         * Look for third '.'.
                                         *
                                         * Set the following:
                                         *
                                         * lbbb has the length of bbb digits.
                                         *
                                         * tempaaa points to where we expect
                                         * the aaa digits to start.
                                         */

        lbbb = (strchr(tempbbb, '.') - tempbbb);
        tempaaa = tempbbb + lbbb + 1;                                   

        if (   (lbbb < 1)
             ||(lbbb > 3)
           )
        {
            printf("Expecting third dot 2 to 4 chars after second dot. Exiting.\n");    
            exit(1);
        }



                                        /* -----------------------------------
                                         * 
                                         * Look for '/'.
                                         *
                                         * Set the following:
                                         *
                                         * laaa has the length of a digits.
                                         *
                                         * tempnn points to where we expect
                                         * the nn digits start.
                                         */

        laaa = (strchr(tempaaa, '/') - tempaaa);
        tempnn = tempaaa + laaa + 1;                                    

        if (   (laaa < 1)
            ||(laaa > 3)
           )
        {
            printf("Expecting slash 2 to 4 chars after third dot. Exiting.\n"); 
            exit(1);
        }


                                        /* -----------------------------------
                                         * 
                                         * Look for space after prefix length.
                                         *
                                         * Set the following:
                                         *
                                         * lnn has the length of the nn digits.
                                         * tempspace to point to that space.
                                         */

        lnn = (strchr(tempnn, ' ') - tempnn);
        tempspace = tempnn + lnn;
        
        if (   (lnn < 1)
             ||(lnn > 2)
           )
        {
            printf("Expecting 1 or 2 digits of prefix length. Exiting.\n"); 
            exit(1);
        }

                                        /* Now convert these digit strings into
                                         * integers.  Note this does not check
                                         * for invalid characters like v in:
                                         *
                                         * "12v.3v.v.0/v4 " 
                                         *
                                         * We are not expecting human-generated
                                         * input - just stuff from a BGP routing
                                         * table.
                                         */
        pfddd = atoi(tempddd);
        pfccc = atoi(tempccc);
        pfbbb = atoi(tempbbb);
        pfaaa = atoi(tempaaa);
        pfnn =  atoi(tempnn);

                                        /* Check the ranges.
                                         */ 

        if (   (pfddd < 0)
             ||(pfddd > 223)
           )
        {
            printf("Expecting ddd in 0 to 223 range. Exiting.\n");  
            exit(1);
        }

        if (   (pfccc < 0)
             ||(pfccc > 255)
           )
        {
            printf("Expecting ccc in 0 to 255 range. Exiting.\n");  
            exit(1);
        }

        if (   (pfbbb < 0)
             ||(pfbbb > 255)
           )
        {
            printf("Expecting bbb in 0 to 255 range. Exiting.\n");  
            exit(1);
        }

        if (   (pfaaa < 0)
             ||(pfaaa > 255)
           )
        {
            printf("Expecting aaa in 0 to 255 range. Exiting.\n");  
            exit(1);
        }

        if (   (pfnn < 8)
             ||(pfnn  > 32)
           )
        {
            printf("Expecting nn in 8 to 32 range. Exiting.\n");    
            exit(1);
        }

                                        /* Debugging line.
                                         */     
/*                                                  
 *  printf("Found: ddd=%d  ccc=%d  bbb=%d  aaa=%d  nn=%d p=%d\n", pfddd, pfccc, pfbbb, pfaaa, pfnn, p);
 */

                                        /* Ignore lines with prefixes longer than /24.
                                         */
                                         
        if (pfnn <=24)
        {                                            
                                        /* Calculate depth to which this prefix
                                         * may cross the boundaries of longer
                                         * prefixes.
                                         *
                                         * Find the lowest bit number which is set.
                                         *
                                         * I would expect that the aaa should be 0,
                                         * because no self-respecting BGP prefix
                                         * shorter than 23 would start on anything
                                         * other than x.x.x.0.  
                                         *
                                         * Spit the dummy (AKA "pacifier") if so.   
                                         */
        
            if (pfaaa !=0)
            {
                printf("Found a prefix of /24 or shorter which starts on a boundary with bits 7 to 0 not all zero. Exiting.\n");    
                exit(1);
            }
            


                                         
            pfaddress =    (pfddd * 256 * 256 * 256)
                         + (pfccc       * 256 * 256)
                         + (pfbbb             * 256);
                         
                                        /* Count the zeroes going left from bit 0,
                                         * by shifting the address to the right 
                                         * until we find a 1 in bit 0.
                                         * 
                                         * Watch out for 0.0.0.0!
                                         */ 
                                         
            lowestbit = 0; 
            while (   ((pfaddress & 1) == 0)
                    &&(lowestbit < 31)
                  )   
            {    
                lowestbit++;
                (pfaddress = pfaddress >> 1);
            }
                                         
            depth = 32 - pfnn - lowestbit;
            
            if (depth < 0) depth = 0;
            
                                        /* Print the output line.
                                         *
                                         * First the address and prefix, in 
                                         * leading zeroes format, so all lines
                                         * have the same format and can later
                                         * be sorted if they get out of order
                                         * as we play with them.
                                         *
                                         * aaa will be zero in all the lines
                                         * we feed to the output file.
                                         *
                                         * tempspace points to the start of the
                                         * remainder of the line which we didn't
                                         * parse.  That has a newline in it.
                                         */         

            fprintf(outfile, "%03d.%03d.%03d.%03d/%02d , D%02d , %s",
                              pfddd,
                              pfccc,
                              pfbbb,
                              pfaaa,
                              pfnn,
                              depth,
                              tempspace);                                                       
        }
    }
}


