/* rand31-park-miller-carta.c          Version 1.00  2005 September 15
 *
 * Fast implementation of the Park Miller (1998) linear congruential 
 * pseudo-random number generator, using David G. Carta's optimisation
 * which needs only 32 bit math, and no division.
 *
 * This file and its .h file is intended to be used in other projects.
 *
 * Public domain 2005 Robin Whittle  rw@firstpr.com.au
 *
 * For a full explanation, latest updates and the history of these
 * algorithms, see http://www.firstpr.com.au/dsp/rand31/
 *
 * Do not strip out or reformat my comments!  One day, some poor sod - 
 * perhaps you or one of your friends, will want to know how the hell 
 * this thing works.
 *
 * See accompanying test program:
 *
 *   rand31-park-miller-carta.c 
 *   rand31-park-miller-carta.h     
 */

#include "rand31-park-miller-carta.h"
#include <math.h>

                                    /* nextrand()
                                     *
                                     * Generate next pseudo random number.
                                     *                                  
                                     * Multiplier constant = 16807 = 7^5.  
                                     * This is 15 bits.
                                     *
                                     * Park and Miller in 1993 recommend
                                     * 48271, which they say produces a 
                                     * somewhat better quality of 
                                     * pseudo-random results.  
                                     *
                                     * 48271 can't be used with the 
                                     * following implementation of Carta's 
                                     * algorithm, since it is 16 bits and 
                                     * would result in bit 31 potentially 
                                     * being set in qlo in the first
                                     * multiplication.  (A similar problem
                                     * occurs later with the higher bits of
                                     * phi.)
                                     * 
                                     * Another way of saying this is that
                                     * we rely on having only 31 bits of
                                     * multiplication product in qlo, 
                                     * because we need to know that this
                                     * result does not exceed 0x7FFFFFFF.
                                     * We need to know this in order for us 
                                     * to know that after adding in 
                                     * something else, that the sum will 
                                     * not exceed 2 * 0x7FFFFFFF.
                                     */
#define consta 16807            

                                    /* Modulus constant = 2^31 - 1 =
                                     * 0x7FFFFFFF.   We use this explicitly
                                     * in the code, rather than define it
                                     * somewhere, because this is a value
                                     * which must not be changed and should
                                     * always be recognised as a zero 
                                     * followed by 31 ones.
                                     */
                                            
long unsigned int nextrand()
{
                                    /* Two 32 bit integers for holding
                                     * parts of the (seed1 * consta)
                                     * multiplication product which would 
                                     * normally need a 46 bit word. 
                                     * 
                                     * qlo 31 bits       30  -  0 
                                     * phi 30 bits   45  -  16  
                                     *
                                     * These overlap in their value.
                                     *
                                     * Bit 0 of phi has the same weight in 
                                     * the result as bit 16 of qlo.
                                     */
                                    
    long unsigned int phi, qlo;

                                    /* qlo = 31 bits:
                                     *  
                                     *    low 16 bits (15-0) of seed31 
                                     *  * 15 bit consta 
                                     */
                                     
    qlo = consta * (seed31 & 0xFFFF);
        
                                    /* phi = 30 bits:
                                     *
                                     *    high 15 bits (30-16) of seed31
                                     *  * 15 bit consta 
                                     */
                                    
    phi = consta * (seed31 >> 16);
        
                                    /* The new pseudo-random number is the 
                                     * 46 bit product mod(0x7FFFFFF).  Our
                                     * task is to calculate it with 32
                                     * bit words and maths, and without
                                     * division.
                                     *
                                     * The first section is easy to
                                     * understand.  We have a bunch of
                                     * bits - bits 14 to 0 of phi - 
                                     * which overlap with and carry the
                                     * same weight as bits 30 to 16 of
                                     * qlo.  We want to add them into
                                     * qlo and then do mod(0x7FFFFFFF)
                                     * before continuing.
                                     *
                                     * Add the low 15 bits of phi into
                                     * bits 30-16 of qlo.  
                                     */
                                    
    qlo += (phi & 0x7FFF) << 16;
        
                                    /* First of two identical modulus 
                                     * operations.
                                     * 
                                     * This algorithm relies on the maximum 
                                     * value in qlo being less than:
                                     *  
                                     *  2 * 0x7FFFFFFF = 0xFFFFFFFE
                                     *
                                     * We can be sure of this because the 
                                     * maximum values which could be added 
                                     *are:
                                     *
                                     *    3322 2222 2222 1111 1111 11
                                     *    1098 7654 3210 9876 5432 1098 7654 3210 
                                     *
                                     *    0111 1111 1111 1111 1111 1111 1111 1111 
                                     * +  0111 1111 1111 1111 0000 0000 0000 0000
                                     * =  1111 1111 1111 1110 1111 1111 1111 1111 
                                     *
                                     * This is lower than 0xFFFFFFFE:
                                     *
                                     *    1111 1111 1111 1111 1111 1111 1111 1111
                                     *
                                     * This means that in doing the modulus, 
                                     * we only need to consider whether qlo 
                                     * is above 0x7FFFFFFF or not.  
                                     * If so, we subtract 0x7FFFFFFF.
                                     *
                                     * One way of doing the test in C or C++ 
                                     * is: 
                                     * 
                                     *      if (qlo > constm)
                                     *
                                     * However, there is another approach 
                                     * suitable for assembler - test bit 32.  
                                     * The dsPIC has a bit test and 
                                     * conditional skip instruction which 
                                     * would be fine for this.
                                     *
                                     * If the value is > 0x7FFFFFF, subtract 
                                     * 0x7FFFFFF. 
                                     *
                                     * To do this subtraction, Ray Gardner 
                                     * used AND 0x7FFFFFFF (which is the same 
                                     * as clearing bit 31) followed by an 
                                     * increment.  This is the same as 
                                     * subtracting 0x80000000 and adding 1.
                                     *
                                     *      if (lo > m)
                                     *      {
                                     *          lo &= m;
                                     *          ++lo;
                                     *      }
                                     * 
                                     * This might be a good approach on a 
                                     * DSP: clear bit 31, leaving the the 31 
                                     * bits 30 to 0 untouched, and then 
                                     * increment the 31 bit result.
                                     *
                                     * (Because we know the value we are
                                     *  working on was less than 2 *
                                     *  0x7FFFFFFF, we know that if bit 31 
                                     *  was set, that the other bits 
                                     *  couldn't all be on.  So we can be 
                                     *  sure the increment won't result in 
                                     *  an overflow to bit 31 and therefore a
                                     *  to a 32 bit result.)
                                     *  
                                     * This may be faster than loading the 
                                     * constant 0x7FFFFFFF and subtracting
                                     * it.
                                     *
                                     * Here, we are writing C or C++ and
                                     * we know the machine handles 32 bit 
                                     * unsigned data. So we simply subtract 
                                     * 0x7FFFFFFF.
                                     */
            
    if (qlo > 0x7FFFFFFF) qlo -= 0x7FFFFFFF;

        
                                    /* So far, we got some of our total
                                     * result in qlo, after doing a 
                                     * mod(0x7FFFFFFF).  So the result 
                                     * contained in qlo and bits 29 to 15 of
                                     * phi is not necessarily exactly the 
                                     * same as the full 46 bit multiplication
                                     * product, since we might have already
                                     * subtracted 0x7FFFFFFF.
                                     *
                                     * The only other part of the result
                                     * we need to deal with is bits
                                     * 29 to 15 of phi. 
                                     *
                                     * These bits carry weights of bits
                                     * 45 to 31 in the value of the 
                                     * multiplication result.
                                     *
                                     * David Carta writes that in order
                                     * to get the mod(0x7FFFFFF) of the
                                     * multiplication product we should
                                     * simply add these bits into the
                                     * bit positions 14 to 0 of qlo.
                                     * His explanation is too long to
                                     * repeat here.  Mine is easy to
                                     * understand. 
                                     *
                                     * Lets define some labels:
                                     *
                                     *  X = 31 bits 30 to 0.
                                     *      The lower 31 bits of the 
                                     *      multiplication product.
                                     *      (qlo contains X, but
                                     *       mod(0x7FFFFFFF). )
                                     *
                                     *  Y = 15 bits 14 to 0.
                                     *      Bits 29 to 15 of phi, ignoring 
                                     *      the lower bits.  So bit 15 of phi
                                     *      is bit 0 of Y.  This means bit 0 
                                     *      of Y has the same weight as would 
                                     *      32nd bit (bit 31) of X.  
                                     *      
                                     *
                                     * If we were doing 46 bit math, the 
                                     * multipication product:
                                     *
                                     *   seed31 * consta would be:
                                     *
                                     *    (Y * 0x80000000) 
                                     *  +  X
                                     *
                                     * Now observe that this is the same as:
                                     *
                                     *    (Y * 0x7FFFFFFF) 
                                     *  + (Y * 0x00000001) 
                                     *  +  X
                                     *                                          
                                     * However, we don't need or want a 46 bit 
                                     * result.  We only want that result
                                     * mod(0x7FFFFFFF).  Therfore we can
                                     * ignore the first line above and use 
                                     * for our result:
                                     *  
                                     *    Y
                                     *  + X 
                                     *
                                     * which we then need to do 
                                     * mod(0x7FFFFFFF) on.
                                     *
                                     *   phi bits 29 to 15 = Y
                                     * 
                                     *   qlo = X mod(7FFFFFFF)
                                     *
                                     * Since we are going to do a modulus
                                     * anyway, qlo is just as good for our
                                     * final result as X.  So we do as
                                     * David Carta suggests and add the 
                                     * 15 upper bits (in what he called 'p'
                                     * but we call "phi bits 29 to 15") 
                                     * to the 15 lower bits (his 'q' but
                                     * we use "qlo") before doing the final
                                     * mod(0x7FFFFFF) operation.
                                     */
    qlo += phi >> 15;

                                    /* In order to be able to get away with
                                     * this, and to perform the same simple
                                     * mod(0x7FFFFFFF) operation as we did
                                     * above, we need to be sure that the
                                     * result of the addition will not 
                                     * exceed:
                                     *                                      
                                     * 2 * 0x7FFFFFFF = 0xFFFFFFFE
                                     * 
                                     * As before, this is easy, since the 
                                     * greatest values of X (qlo) and Y
                                     * (shifted bits from phi) are as 
                                     * follows.  We are adding in 15 bits
                                     * from phi, but this time in lower
                                     * positions.
                                     *  
                                     *    3322 2222 2222 1111 1111 11
                                     *    1098 7654 3210 9876 5432 1098 7654 3210 
                                     *
                                     *    0111 1111 1111 1111 1111 1111 1111 1111
                                     * +  0000 0000 0000 0000 0111 1111 1111 1111 
                                     * =  1000 0000 0000 0000 0111 1111 1111 1110
                                     *
                                     * This is less than FFFFFFFE:
                                     *
                                     *    1111 1111 1111 1111 1111 1111 1111 1110   
                                     *
                                     * Then do the mod(0x7FFFFFFF) as before.
                                     *
                                     * Note that in the vast majority of cases,
                                     * qlo will be less than 0x7FFFFFFFF.   
                                     */
                                    
    if (qlo > 0x7FFFFFFF) qlo -= 0x7FFFFFFF;        
        
                                    /* qlo contains the new pseudo-random
                                     * number.  Store it to the seed31 and
                                     * return it.
                                     */
        
    return ( seed31 = (long)qlo );      
}


/*---------------------------------------------------------------------------*/

                                    /* seedi()
                                     *
                                     * Set the seed from a long unsigned 
                                     * integer.  If zero is used, then
                                     * the seed will be set to 1.
                                     */
                                                                        
void seedi(long unsigned int seedin)
{
    if (seedin == 0) seedin = 1;
    seed31 = seedin;
}
                                    
                                    /* ranlui()
                                     *
                                     * Return next pseudo-random value as
                                     * a long unsigned integer.
                                     */
                                    
long unsigned int ranlui(void)  
{
    return nextrand();
}

                                    // ranf()
                                    //
                                    // Return next pseudo-random value as
                                    // a floating point value.

float ranf(void)  
{
    return (nextrand() / 2147483647.0 );
}

